From 9fd392404a41581a66d600f249b7d523fc75262b Mon Sep 17 00:00:00 2001
From: Bertrand Drouvot <bertranddrouvot.pg@gmail.com>
Date: Wed, 8 Nov 2023 09:28:01 +0000
Subject: [PATCH v11] split tables and indexes stats

---
 src/backend/access/common/relation.c          |  10 +-
 src/backend/access/index/indexam.c            |   2 +-
 src/backend/catalog/heap.c                    |   7 +-
 src/backend/catalog/index.c                   |   4 +-
 src/backend/catalog/system_views.sql          | 112 +++---
 src/backend/storage/buffer/bufmgr.c           |  12 +-
 src/backend/utils/activity/Makefile           |   3 +-
 src/backend/utils/activity/meson.build        |   3 +-
 src/backend/utils/activity/pgstat.c           |  31 +-
 src/backend/utils/activity/pgstat_index.c     | 312 ++++++++++++++++
 .../{pgstat_relation.c => pgstat_table.c}     | 284 +++++++--------
 src/backend/utils/activity/pgstat_xact.c      |   8 +-
 src/backend/utils/adt/pgstatfuncs.c           | 223 ++++++++----
 src/backend/utils/cache/relcache.c            |  13 +-
 src/include/catalog/pg_proc.dat               | 172 +++++----
 src/include/pgstat.h                          | 171 +++++++--
 src/include/utils/pgstat_internal.h           |  31 +-
 src/include/utils/rel.h                       |   5 +-
 src/test/isolation/expected/stats.out         | 304 ++++++++--------
 src/test/isolation/expected/stats_1.out       | 266 +++++++-------
 src/test/isolation/specs/stats.spec           |  16 +-
 src/test/recovery/t/029_stats_restart.pl      |  23 +-
 .../recovery/t/030_stats_cleanup_replica.pl   |   2 +-
 src/test/regress/expected/rules.out           |  96 ++---
 src/test/regress/expected/stats.out           | 333 ++++++++++++------
 src/test/regress/sql/stats.sql                | 143 ++++++--
 src/tools/pgindent/typedefs.list              |   4 +-
 27 files changed, 1666 insertions(+), 924 deletions(-)
   6.3% src/backend/catalog/
  19.1% src/backend/utils/activity/
   6.8% src/backend/utils/adt/
   8.1% src/include/catalog/
   7.0% src/include/
  27.3% src/test/isolation/expected/
  16.4% src/test/regress/expected/
   5.8% src/test/regress/sql/

diff --git a/src/backend/access/common/relation.c b/src/backend/access/common/relation.c
index 4017e175e3..fca166a063 100644
--- a/src/backend/access/common/relation.c
+++ b/src/backend/access/common/relation.c
@@ -73,7 +73,10 @@ relation_open(Oid relationId, LOCKMODE lockmode)
 	if (RelationUsesLocalBuffers(r))
 		MyXactFlags |= XACT_FLAGS_ACCESSEDTEMPNAMESPACE;
 
-	pgstat_init_relation(r);
+	if (r->rd_rel->relkind == RELKIND_INDEX)
+		pgstat_init_index(r);
+	else
+		pgstat_init_table(r);
 
 	return r;
 }
@@ -123,7 +126,10 @@ try_relation_open(Oid relationId, LOCKMODE lockmode)
 	if (RelationUsesLocalBuffers(r))
 		MyXactFlags |= XACT_FLAGS_ACCESSEDTEMPNAMESPACE;
 
-	pgstat_init_relation(r);
+	if (r->rd_rel->relkind == RELKIND_INDEX)
+		pgstat_init_index(r);
+	else
+		pgstat_init_table(r);
 
 	return r;
 }
diff --git a/src/backend/access/index/indexam.c b/src/backend/access/index/indexam.c
index b25b03f7ab..5e4722b763 100644
--- a/src/backend/access/index/indexam.c
+++ b/src/backend/access/index/indexam.c
@@ -590,7 +590,7 @@ index_fetch_heap(IndexScanDesc scan, TupleTableSlot *slot)
 									&scan->xs_heap_continue, &all_dead);
 
 	if (found)
-		pgstat_count_heap_fetch(scan->indexRelation);
+		pgstat_count_index_fetch(scan->indexRelation);
 
 	/*
 	 * If we scanned a whole HOT chain and found only dead tuples, tell index
diff --git a/src/backend/catalog/heap.c b/src/backend/catalog/heap.c
index d03c961678..e9f8d0fb2f 100644
--- a/src/backend/catalog/heap.c
+++ b/src/backend/catalog/heap.c
@@ -404,7 +404,10 @@ heap_create(const char *relname,
 									 reltablespace);
 
 	/* ensure that stats are dropped if transaction aborts */
-	pgstat_create_relation(rel);
+	if (rel->rd_rel->relkind == RELKIND_INDEX)
+		pgstat_create_index(rel);
+	else
+		pgstat_create_table(rel);
 
 	return rel;
 }
@@ -1840,7 +1843,7 @@ heap_drop_with_catalog(Oid relid)
 		RelationDropStorage(rel);
 
 	/* ensure that stats are dropped if transaction commits */
-	pgstat_drop_relation(rel);
+	pgstat_drop_table(rel);
 
 	/*
 	 * Close relcache entry, but *keep* AccessExclusiveLock on the relation
diff --git a/src/backend/catalog/index.c b/src/backend/catalog/index.c
index 143fae01eb..7ae2719563 100644
--- a/src/backend/catalog/index.c
+++ b/src/backend/catalog/index.c
@@ -1757,7 +1757,7 @@ index_concurrently_swap(Oid newIndexId, Oid oldIndexId, const char *oldName)
 	changeDependenciesOn(RelationRelationId, oldIndexId, newIndexId);
 
 	/* copy over statistics from old to new index */
-	pgstat_copy_relation_stats(newClassRel, oldClassRel);
+	pgstat_copy_index_stats(newClassRel, oldClassRel);
 
 	/* Copy data of pg_statistic from the old index to the new one */
 	CopyStatistics(oldIndexId, newIndexId);
@@ -2331,7 +2331,7 @@ index_drop(Oid indexId, bool concurrent, bool concurrent_lock_mode)
 		RelationDropStorage(userIndexRelation);
 
 	/* ensure that stats are dropped if transaction commits */
-	pgstat_drop_relation(userIndexRelation);
+	pgstat_drop_index(userIndexRelation);
 
 	/*
 	 * Close and flush the index's relcache entry, to ensure relcache doesn't
diff --git a/src/backend/catalog/system_views.sql b/src/backend/catalog/system_views.sql
index b65f6b5249..ab36e1adca 100644
--- a/src/backend/catalog/system_views.sql
+++ b/src/backend/catalog/system_views.sql
@@ -654,30 +654,30 @@ CREATE VIEW pg_stat_all_tables AS
             C.oid AS relid,
             N.nspname AS schemaname,
             C.relname AS relname,
-            pg_stat_get_numscans(C.oid) AS seq_scan,
-            pg_stat_get_lastscan(C.oid) AS last_seq_scan,
-            pg_stat_get_tuples_returned(C.oid) AS seq_tup_read,
-            sum(pg_stat_get_numscans(I.indexrelid))::bigint AS idx_scan,
-            max(pg_stat_get_lastscan(I.indexrelid)) AS last_idx_scan,
-            sum(pg_stat_get_tuples_fetched(I.indexrelid))::bigint +
-            pg_stat_get_tuples_fetched(C.oid) AS idx_tup_fetch,
-            pg_stat_get_tuples_inserted(C.oid) AS n_tup_ins,
-            pg_stat_get_tuples_updated(C.oid) AS n_tup_upd,
-            pg_stat_get_tuples_deleted(C.oid) AS n_tup_del,
-            pg_stat_get_tuples_hot_updated(C.oid) AS n_tup_hot_upd,
-            pg_stat_get_tuples_newpage_updated(C.oid) AS n_tup_newpage_upd,
-            pg_stat_get_live_tuples(C.oid) AS n_live_tup,
-            pg_stat_get_dead_tuples(C.oid) AS n_dead_tup,
-            pg_stat_get_mod_since_analyze(C.oid) AS n_mod_since_analyze,
-            pg_stat_get_ins_since_vacuum(C.oid) AS n_ins_since_vacuum,
-            pg_stat_get_last_vacuum_time(C.oid) as last_vacuum,
-            pg_stat_get_last_autovacuum_time(C.oid) as last_autovacuum,
-            pg_stat_get_last_analyze_time(C.oid) as last_analyze,
-            pg_stat_get_last_autoanalyze_time(C.oid) as last_autoanalyze,
-            pg_stat_get_vacuum_count(C.oid) AS vacuum_count,
-            pg_stat_get_autovacuum_count(C.oid) AS autovacuum_count,
-            pg_stat_get_analyze_count(C.oid) AS analyze_count,
-            pg_stat_get_autoanalyze_count(C.oid) AS autoanalyze_count
+            pg_stat_get_tab_numscans(C.oid) AS seq_scan,
+            pg_stat_get_tab_lastscan(C.oid) AS last_seq_scan,
+            pg_stat_get_tab_tuples_returned(C.oid) AS seq_tup_read,
+            sum(pg_stat_get_ind_numscans(I.indexrelid))::bigint AS idx_scan,
+            max(pg_stat_get_ind_lastscan(I.indexrelid)) AS last_idx_scan,
+            sum(pg_stat_get_ind_tuples_fetched(I.indexrelid))::bigint +
+            pg_stat_get_tab_tuples_fetched(C.oid) AS idx_tup_fetch,
+            pg_stat_get_tab_tuples_inserted(C.oid) AS n_tup_ins,
+            pg_stat_get_tab_tuples_updated(C.oid) AS n_tup_upd,
+            pg_stat_get_tab_tuples_deleted(C.oid) AS n_tup_del,
+            pg_stat_get_tab_tuples_hot_updated(C.oid) AS n_tup_hot_upd,
+            pg_stat_get_tab_tuples_newpage_updated(C.oid) AS n_tup_newpage_upd,
+            pg_stat_get_tab_live_tuples(C.oid) AS n_live_tup,
+            pg_stat_get_tab_dead_tuples(C.oid) AS n_dead_tup,
+            pg_stat_get_tab_mod_since_analyze(C.oid) AS n_mod_since_analyze,
+            pg_stat_get_tab_ins_since_vacuum(C.oid) AS n_ins_since_vacuum,
+            pg_stat_get_tab_last_vacuum_time(C.oid) as last_vacuum,
+            pg_stat_get_tab_last_autovacuum_time(C.oid) as last_autovacuum,
+            pg_stat_get_tab_last_analyze_time(C.oid) as last_analyze,
+            pg_stat_get_tab_last_autoanalyze_time(C.oid) as last_autoanalyze,
+            pg_stat_get_tab_vacuum_count(C.oid) AS vacuum_count,
+            pg_stat_get_tab_autovacuum_count(C.oid) AS autovacuum_count,
+            pg_stat_get_tab_analyze_count(C.oid) AS analyze_count,
+            pg_stat_get_tab_autoanalyze_count(C.oid) AS autoanalyze_count
     FROM pg_class C LEFT JOIN
          pg_index I ON C.oid = I.indrelid
          LEFT JOIN pg_namespace N ON (N.oid = C.relnamespace)
@@ -689,16 +689,16 @@ CREATE VIEW pg_stat_xact_all_tables AS
             C.oid AS relid,
             N.nspname AS schemaname,
             C.relname AS relname,
-            pg_stat_get_xact_numscans(C.oid) AS seq_scan,
-            pg_stat_get_xact_tuples_returned(C.oid) AS seq_tup_read,
-            sum(pg_stat_get_xact_numscans(I.indexrelid))::bigint AS idx_scan,
-            sum(pg_stat_get_xact_tuples_fetched(I.indexrelid))::bigint +
-            pg_stat_get_xact_tuples_fetched(C.oid) AS idx_tup_fetch,
-            pg_stat_get_xact_tuples_inserted(C.oid) AS n_tup_ins,
-            pg_stat_get_xact_tuples_updated(C.oid) AS n_tup_upd,
-            pg_stat_get_xact_tuples_deleted(C.oid) AS n_tup_del,
-            pg_stat_get_xact_tuples_hot_updated(C.oid) AS n_tup_hot_upd,
-            pg_stat_get_xact_tuples_newpage_updated(C.oid) AS n_tup_newpage_upd
+            pg_stat_get_tab_xact_numscans(C.oid) AS seq_scan,
+            pg_stat_get_tab_xact_tuples_returned(C.oid) AS seq_tup_read,
+            sum(pg_stat_get_ind_xact_numscans(I.indexrelid))::bigint AS idx_scan,
+            sum(pg_stat_get_ind_xact_tuples_fetched(I.indexrelid))::bigint +
+            pg_stat_get_tab_xact_tuples_fetched(C.oid) AS idx_tup_fetch,
+            pg_stat_get_tab_xact_tuples_inserted(C.oid) AS n_tup_ins,
+            pg_stat_get_tab_xact_tuples_updated(C.oid) AS n_tup_upd,
+            pg_stat_get_tab_xact_tuples_deleted(C.oid) AS n_tup_del,
+            pg_stat_get_tab_xact_tuples_hot_updated(C.oid) AS n_tup_hot_upd,
+            pg_stat_get_tab_xact_tuples_newpage_updated(C.oid) AS n_tup_newpage_upd
     FROM pg_class C LEFT JOIN
          pg_index I ON C.oid = I.indrelid
          LEFT JOIN pg_namespace N ON (N.oid = C.relnamespace)
@@ -730,31 +730,31 @@ CREATE VIEW pg_statio_all_tables AS
             C.oid AS relid,
             N.nspname AS schemaname,
             C.relname AS relname,
-            pg_stat_get_blocks_fetched(C.oid) -
-                    pg_stat_get_blocks_hit(C.oid) AS heap_blks_read,
-            pg_stat_get_blocks_hit(C.oid) AS heap_blks_hit,
+            pg_stat_get_tab_blocks_fetched(C.oid) -
+                    pg_stat_get_tab_blocks_hit(C.oid) AS heap_blks_read,
+            pg_stat_get_tab_blocks_hit(C.oid) AS heap_blks_hit,
             I.idx_blks_read AS idx_blks_read,
             I.idx_blks_hit AS idx_blks_hit,
-            pg_stat_get_blocks_fetched(T.oid) -
-                    pg_stat_get_blocks_hit(T.oid) AS toast_blks_read,
-            pg_stat_get_blocks_hit(T.oid) AS toast_blks_hit,
+            pg_stat_get_tab_blocks_fetched(T.oid) -
+                    pg_stat_get_tab_blocks_hit(T.oid) AS toast_blks_read,
+            pg_stat_get_tab_blocks_hit(T.oid) AS toast_blks_hit,
             X.idx_blks_read AS tidx_blks_read,
             X.idx_blks_hit AS tidx_blks_hit
     FROM pg_class C LEFT JOIN
             pg_class T ON C.reltoastrelid = T.oid
             LEFT JOIN pg_namespace N ON (N.oid = C.relnamespace)
             LEFT JOIN LATERAL (
-              SELECT sum(pg_stat_get_blocks_fetched(indexrelid) -
-                         pg_stat_get_blocks_hit(indexrelid))::bigint
+              SELECT sum(pg_stat_get_ind_blocks_fetched(indexrelid) -
+                         pg_stat_get_ind_blocks_hit(indexrelid))::bigint
                      AS idx_blks_read,
-                     sum(pg_stat_get_blocks_hit(indexrelid))::bigint
+                     sum(pg_stat_get_ind_blocks_hit(indexrelid))::bigint
                      AS idx_blks_hit
               FROM pg_index WHERE indrelid = C.oid ) I ON true
             LEFT JOIN LATERAL (
-              SELECT sum(pg_stat_get_blocks_fetched(indexrelid) -
-                         pg_stat_get_blocks_hit(indexrelid))::bigint
+              SELECT sum(pg_stat_get_ind_blocks_fetched(indexrelid) -
+                         pg_stat_get_ind_blocks_hit(indexrelid))::bigint
                      AS idx_blks_read,
-                     sum(pg_stat_get_blocks_hit(indexrelid))::bigint
+                     sum(pg_stat_get_ind_blocks_hit(indexrelid))::bigint
                      AS idx_blks_hit
               FROM pg_index WHERE indrelid = T.oid ) X ON true
     WHERE C.relkind IN ('r', 't', 'm');
@@ -776,10 +776,10 @@ CREATE VIEW pg_stat_all_indexes AS
             N.nspname AS schemaname,
             C.relname AS relname,
             I.relname AS indexrelname,
-            pg_stat_get_numscans(I.oid) AS idx_scan,
-            pg_stat_get_lastscan(I.oid) AS last_idx_scan,
-            pg_stat_get_tuples_returned(I.oid) AS idx_tup_read,
-            pg_stat_get_tuples_fetched(I.oid) AS idx_tup_fetch
+            pg_stat_get_ind_numscans(I.oid) AS idx_scan,
+            pg_stat_get_ind_lastscan(I.oid) AS last_idx_scan,
+            pg_stat_get_ind_tuples_returned(I.oid) AS idx_tup_read,
+            pg_stat_get_ind_tuples_fetched(I.oid) AS idx_tup_fetch
     FROM pg_class C JOIN
             pg_index X ON C.oid = X.indrelid JOIN
             pg_class I ON I.oid = X.indexrelid
@@ -803,9 +803,9 @@ CREATE VIEW pg_statio_all_indexes AS
             N.nspname AS schemaname,
             C.relname AS relname,
             I.relname AS indexrelname,
-            pg_stat_get_blocks_fetched(I.oid) -
-                    pg_stat_get_blocks_hit(I.oid) AS idx_blks_read,
-            pg_stat_get_blocks_hit(I.oid) AS idx_blks_hit
+            pg_stat_get_ind_blocks_fetched(I.oid) -
+                    pg_stat_get_ind_blocks_hit(I.oid) AS idx_blks_read,
+            pg_stat_get_ind_blocks_hit(I.oid) AS idx_blks_hit
     FROM pg_class C JOIN
             pg_index X ON C.oid = X.indrelid JOIN
             pg_class I ON I.oid = X.indexrelid
@@ -827,9 +827,9 @@ CREATE VIEW pg_statio_all_sequences AS
             C.oid AS relid,
             N.nspname AS schemaname,
             C.relname AS relname,
-            pg_stat_get_blocks_fetched(C.oid) -
-                    pg_stat_get_blocks_hit(C.oid) AS blks_read,
-            pg_stat_get_blocks_hit(C.oid) AS blks_hit
+            pg_stat_get_ind_blocks_fetched(C.oid) -
+                    pg_stat_get_tab_blocks_hit(C.oid) AS blks_read,
+            pg_stat_get_tab_blocks_hit(C.oid) AS blks_hit
     FROM pg_class C
             LEFT JOIN pg_namespace N ON (N.oid = C.relnamespace)
     WHERE C.relkind = 'S';
diff --git a/src/backend/storage/buffer/bufmgr.c b/src/backend/storage/buffer/bufmgr.c
index f7c67d504c..bc2ddee0a3 100644
--- a/src/backend/storage/buffer/bufmgr.c
+++ b/src/backend/storage/buffer/bufmgr.c
@@ -799,11 +799,19 @@ ReadBufferExtended(Relation reln, ForkNumber forkNum, BlockNumber blockNum,
 	 * Read the buffer, and update pgstat counters to reflect a cache hit or
 	 * miss.
 	 */
-	pgstat_count_buffer_read(reln);
+	if (reln->rd_rel->relkind == RELKIND_INDEX)
+		pgstat_count_index_buffer_read(reln);
+	else
+		pgstat_count_table_buffer_read(reln);
 	buf = ReadBuffer_common(RelationGetSmgr(reln), reln->rd_rel->relpersistence,
 							forkNum, blockNum, mode, strategy, &hit);
 	if (hit)
-		pgstat_count_buffer_hit(reln);
+	{
+		if (reln->rd_rel->relkind == RELKIND_INDEX)
+			pgstat_count_index_buffer_hit(reln);
+		else
+			pgstat_count_table_buffer_hit(reln);
+	}
 	return buf;
 }
 
diff --git a/src/backend/utils/activity/Makefile b/src/backend/utils/activity/Makefile
index a4397ef907..341a41f2b4 100644
--- a/src/backend/utils/activity/Makefile
+++ b/src/backend/utils/activity/Makefile
@@ -25,11 +25,12 @@ OBJS = \
 	pgstat_database.o \
 	pgstat_function.o \
 	pgstat_io.o \
-	pgstat_relation.o \
+	pgstat_index.o \
 	pgstat_replslot.o \
 	pgstat_shmem.o \
 	pgstat_slru.o \
 	pgstat_subscription.o \
+	pgstat_table.o \
 	pgstat_wal.o \
 	pgstat_xact.o \
 	wait_event.o \
diff --git a/src/backend/utils/activity/meson.build b/src/backend/utils/activity/meson.build
index 46a27e7548..9248a79941 100644
--- a/src/backend/utils/activity/meson.build
+++ b/src/backend/utils/activity/meson.build
@@ -10,11 +10,12 @@ backend_sources += files(
   'pgstat_database.c',
   'pgstat_function.c',
   'pgstat_io.c',
-  'pgstat_relation.c',
+  'pgstat_index.c',
   'pgstat_replslot.c',
   'pgstat_shmem.c',
   'pgstat_slru.c',
   'pgstat_subscription.c',
+  'pgstat_table.c',
   'pgstat_wal.c',
   'pgstat_xact.c',
 )
diff --git a/src/backend/utils/activity/pgstat.c b/src/backend/utils/activity/pgstat.c
index d743fc0b28..ebace8c998 100644
--- a/src/backend/utils/activity/pgstat.c
+++ b/src/backend/utils/activity/pgstat.c
@@ -73,10 +73,11 @@
  * - pgstat_database.c
  * - pgstat_function.c
  * - pgstat_io.c
- * - pgstat_relation.c
+ * - pgstat_index.c
  * - pgstat_replslot.c
  * - pgstat_slru.c
  * - pgstat_subscription.c
+ * - pgstat_table.c
  * - pgstat_wal.c
  *
  * Whenever possible infrastructure files should not contain code related to
@@ -277,18 +278,32 @@ static const PgStat_KindInfo pgstat_kind_infos[PGSTAT_NUM_KINDS] = {
 		.reset_timestamp_cb = pgstat_database_reset_timestamp_cb,
 	},
 
-	[PGSTAT_KIND_RELATION] = {
-		.name = "relation",
+	[PGSTAT_KIND_TABLE] = {
+		.name = "table",
 
 		.fixed_amount = false,
 
-		.shared_size = sizeof(PgStatShared_Relation),
-		.shared_data_off = offsetof(PgStatShared_Relation, stats),
-		.shared_data_len = sizeof(((PgStatShared_Relation *) 0)->stats),
+		.shared_size = sizeof(PgStatShared_Table),
+		.shared_data_off = offsetof(PgStatShared_Table, stats),
+		.shared_data_len = sizeof(((PgStatShared_Table *) 0)->stats),
 		.pending_size = sizeof(PgStat_TableStatus),
 
-		.flush_pending_cb = pgstat_relation_flush_cb,
-		.delete_pending_cb = pgstat_relation_delete_pending_cb,
+		.flush_pending_cb = pgstat_table_flush_cb,
+		.delete_pending_cb = pgstat_table_delete_pending_cb,
+	},
+
+	[PGSTAT_KIND_INDEX] = {
+		.name = "index",
+
+		.fixed_amount = false,
+
+		.shared_size = sizeof(PgStatShared_Index),
+		.shared_data_off = offsetof(PgStatShared_Index, stats),
+		.shared_data_len = sizeof(((PgStatShared_Index *) 0)->stats),
+		.pending_size = sizeof(PgStat_IndexStatus),
+
+		.flush_pending_cb = pgstat_index_flush_cb,
+		.delete_pending_cb = pgstat_index_delete_pending_cb,
 	},
 
 	[PGSTAT_KIND_FUNCTION] = {
diff --git a/src/backend/utils/activity/pgstat_index.c b/src/backend/utils/activity/pgstat_index.c
new file mode 100644
index 0000000000..131d431a30
--- /dev/null
+++ b/src/backend/utils/activity/pgstat_index.c
@@ -0,0 +1,312 @@
+/* -------------------------------------------------------------------------
+ *
+ * pgstat_index.c
+ *	  Implementation of index statistics.
+ *
+ * This file contains the implementation of function index. It is kept
+ * separate from pgstat.c to enforce the line between the statistics access /
+ * storage implementation and the details about individual types of
+ * statistics.
+ *
+ * Copyright (c) 2001-2023, PostgreSQL Global Development Group
+ *
+ * IDENTIFICATION
+ *	  src/backend/utils/activity/pgstat_index.c
+ * -------------------------------------------------------------------------
+ */
+
+#include "postgres.h"
+
+#include "access/xact.h"
+#include "utils/pgstat_internal.h"
+#include "utils/rel.h"
+#include "catalog/catalog.h"
+
+static PgStat_IndexStatus *pgstat_prep_index_pending(Oid rel_id, bool isshared);
+
+/*
+ * Copy stats between indexes. This is used for things like REINDEX
+ * CONCURRENTLY.
+ */
+void
+pgstat_copy_index_stats(Relation dst, Relation src)
+{
+	PgStat_StatIndEntry *srcstats;
+	PgStatShared_Index *dstshstats;
+	PgStat_EntryRef *dst_ref;
+
+	srcstats = pgstat_fetch_stat_indentry_ext(src->rd_rel->relisshared,
+											  RelationGetRelid(src));
+	if (!srcstats)
+		return;
+
+	dst_ref = pgstat_get_entry_ref_locked(PGSTAT_KIND_INDEX,
+										  dst->rd_rel->relisshared ? InvalidOid : MyDatabaseId,
+										  RelationGetRelid(dst),
+										  false);
+
+	dstshstats = (PgStatShared_Index *) dst_ref->shared_stats;
+	dstshstats->stats = *srcstats;
+
+	pgstat_unlock_entry(dst_ref);
+}
+
+/*
+ * Initialize a relcache entry to count access statistics.  Called whenever an
+ * index is opened.
+ *
+ * We assume that a relcache entry's pgstatind_info field is zeroed by relcache.c
+ * when the relcache entry is made; thereafter it is long-lived data.
+ *
+ * This does not create a reference to a stats entry in shared memory, nor
+ * allocate memory for the pending stats. That happens in
+ * pgstat_assoc_index().
+ */
+void
+pgstat_init_index(Relation rel)
+{
+	/*
+	 * We only count stats for indexes
+	 */
+	Assert(rel->rd_rel->relkind == RELKIND_INDEX);
+
+	if (!pgstat_track_counts)
+	{
+		if (rel->pgstatind_info != NULL)
+			pgstat_unlink_index(rel);
+
+		/* We're not counting at all */
+		rel->pgstat_enabled = false;
+		rel->pgstatind_info = NULL;
+		return;
+	}
+
+	rel->pgstat_enabled = true;
+}
+
+/*
+ * Prepare for statistics for this index to be collected.
+ *
+ * This ensures we have a reference to the stats entry before stats can be
+ * generated. That is important because an index drop in another
+ * connection could otherwise lead to the stats entry being dropped, which then
+ * later would get recreated when flushing stats.
+ *
+ * This is separate from pgstat_init_index() as it is not uncommon for
+ * relcache entries to be opened without ever getting stats reported.
+ */
+void
+pgstat_assoc_index(Relation rel)
+{
+	Assert(rel->pgstat_enabled);
+	Assert(rel->pgstatind_info == NULL);
+
+	/* Else find or make the PgStat_IndexStatus entry, and update link */
+	rel->pgstatind_info = pgstat_prep_index_pending(RelationGetRelid(rel),
+													rel->rd_rel->relisshared);
+
+	/* don't allow link a stats to multiple relcache entries */
+	Assert(rel->pgstatind_info->relation == NULL);
+
+	/* mark this relation as the owner */
+	rel->pgstatind_info->relation = rel;
+}
+
+/*
+ * Break the mutual link between a relcache entry and pending index stats entry.
+ * This must be called whenever one end of the link is removed.
+ */
+void
+pgstat_unlink_index(Relation rel)
+{
+
+	if (rel->pgstatind_info == NULL)
+		return;
+
+	/* link sanity check for the index stats */
+	if (rel->pgstatind_info)
+	{
+		Assert(rel->pgstatind_info->relation == rel);
+		rel->pgstatind_info->relation = NULL;
+		rel->pgstatind_info = NULL;
+	}
+}
+
+/*
+ * Ensure that index stats are dropped if transaction aborts.
+ */
+void
+pgstat_create_index(Relation rel)
+{
+	pgstat_create_transactional(PGSTAT_KIND_INDEX,
+								rel->rd_rel->relisshared ? InvalidOid : MyDatabaseId,
+								RelationGetRelid(rel));
+}
+
+/*
+ * Ensure that index stats are dropped if transaction commits.
+ */
+void
+pgstat_drop_index(Relation rel)
+{
+	pgstat_drop_transactional(PGSTAT_KIND_INDEX,
+							  rel->rd_rel->relisshared ? InvalidOid : MyDatabaseId,
+							  RelationGetRelid(rel));
+}
+
+/*
+ * Support function for the SQL-callable pgstat* functions. Returns
+ * the collected statistics for one index or NULL. NULL doesn't mean
+ * that the index doesn't exist, just that there are no statistics, so the
+ * caller is better off to report ZERO instead.
+ */
+PgStat_StatIndEntry *
+pgstat_fetch_stat_indentry(Oid relid)
+{
+	return pgstat_fetch_stat_indentry_ext(IsSharedRelation(relid), relid);
+}
+
+/*
+ * More efficient version of pgstat_fetch_stat_indentry(), allowing to specify
+ * whether the to-be-accessed index is shared or not.
+ */
+PgStat_StatIndEntry *
+pgstat_fetch_stat_indentry_ext(bool shared, Oid reloid)
+{
+	Oid			dboid = (shared ? InvalidOid : MyDatabaseId);
+
+	return (PgStat_StatIndEntry *)
+		pgstat_fetch_entry(PGSTAT_KIND_INDEX, dboid, reloid);
+}
+
+/*
+ * find any existing PgStat_IndexStatus entry for rel
+ *
+ * Find any existing PgStat_IndexStatus entry for rel_id in the current
+ * database. If not found, try finding from shared indexes.
+ *
+ * If no entry found, return NULL, don't create a new one
+ */
+PgStat_IndexStatus *
+find_indstat_entry(Oid rel_id)
+{
+	PgStat_EntryRef *entry_ref;
+	PgStat_IndexStatus *indentry = NULL;
+	PgStat_IndexStatus *indexstatus = NULL;
+
+	entry_ref = pgstat_fetch_pending_entry(PGSTAT_KIND_INDEX, MyDatabaseId, rel_id);
+	if (!entry_ref)
+	{
+		entry_ref = pgstat_fetch_pending_entry(PGSTAT_KIND_INDEX, InvalidOid, rel_id);
+		if (!entry_ref)
+			return indexstatus;
+	}
+
+	indentry = (PgStat_IndexStatus *) entry_ref->pending;
+	indexstatus = palloc(sizeof(PgStat_IndexStatus));
+	*indexstatus = *indentry;
+
+	/*
+	 * Reset indexstatus->trans in the copy of PgStat_IndexStatus as it may
+	 * point to a shared memory area.  Its data is saved below, so removing it
+	 * does not matter.
+	 */
+	indexstatus->trans = NULL;
+
+	return indexstatus;
+}
+
+/*
+ * Flush out pending stats for the entry
+ *
+ * If nowait is true, this function returns false if lock could not
+ * immediately acquired, otherwise true is returned.
+ *
+ * Some of the stats are copied to the corresponding pending database stats
+ * entry when successfully flushing.
+ */
+bool
+pgstat_index_flush_cb(PgStat_EntryRef *entry_ref, bool nowait)
+{
+	static const PgStat_IndexCounts all_zeroes;
+	Oid			dboid;
+
+	PgStat_IndexStatus *lstats; /* pending stats entry  */
+	PgStatShared_Index *shrelcomstats;
+	PgStat_StatIndEntry *indentry;	/* index entry of shared stats */
+	PgStat_StatDBEntry *dbentry;	/* pending database entry */
+
+	dboid = entry_ref->shared_entry->key.dboid;
+	lstats = (PgStat_IndexStatus *) entry_ref->pending;
+	shrelcomstats = (PgStatShared_Index *) entry_ref->shared_stats;
+
+	/*
+	 * Ignore entries that didn't accumulate any actual counts, such as
+	 * indexes that were opened by the planner but not used.
+	 */
+	if (memcmp(&lstats->counts, &all_zeroes,
+			   sizeof(PgStat_IndexCounts)) == 0)
+	{
+		return true;
+	}
+
+	if (!pgstat_lock_entry(entry_ref, nowait))
+		return false;
+
+	/* add the values to the shared entry. */
+	indentry = &shrelcomstats->stats;
+
+	indentry->numscans += lstats->counts.numscans;
+
+	if (lstats->counts.numscans)
+	{
+		TimestampTz t = GetCurrentTransactionStopTimestamp();
+
+		if (t > indentry->lastscan)
+			indentry->lastscan = t;
+	}
+	indentry->tuples_returned += lstats->counts.tuples_returned;
+	indentry->tuples_fetched += lstats->counts.tuples_fetched;
+	indentry->blocks_fetched += lstats->counts.blocks_fetched;
+	indentry->blocks_hit += lstats->counts.blocks_hit;
+
+	pgstat_unlock_entry(entry_ref);
+
+	/* The entry was successfully flushed, add the same to database stats */
+	dbentry = pgstat_prep_database_pending(dboid);
+	dbentry->tuples_returned += lstats->counts.tuples_returned;
+	dbentry->tuples_fetched += lstats->counts.tuples_fetched;
+	dbentry->blocks_fetched += lstats->counts.blocks_fetched;
+	dbentry->blocks_hit += lstats->counts.blocks_hit;
+
+	return true;
+}
+
+void
+pgstat_index_delete_pending_cb(PgStat_EntryRef *entry_ref)
+{
+	PgStat_IndexStatus *pending = (PgStat_IndexStatus *) entry_ref->pending;
+
+	if (pending->relation)
+		pgstat_unlink_index(pending->relation);
+}
+
+/*
+ * Find or create a PgStat_IndexStatus entry for rel. New entry is created and
+ * initialized if not exists.
+ */
+static PgStat_IndexStatus *
+pgstat_prep_index_pending(Oid rel_id, bool isshared)
+{
+	PgStat_EntryRef *entry_ref;
+	PgStat_IndexStatus *pending;
+
+	entry_ref = pgstat_prep_pending_entry(PGSTAT_KIND_INDEX,
+										  isshared ? InvalidOid : MyDatabaseId,
+										  rel_id, NULL);
+	pending = entry_ref->pending;
+	pending->r_id = rel_id;
+	pending->r_shared = isshared;
+
+	return pending;
+}
diff --git a/src/backend/utils/activity/pgstat_relation.c b/src/backend/utils/activity/pgstat_table.c
similarity index 76%
rename from src/backend/utils/activity/pgstat_relation.c
rename to src/backend/utils/activity/pgstat_table.c
index f5d726e292..96f6ff67e2 100644
--- a/src/backend/utils/activity/pgstat_relation.c
+++ b/src/backend/utils/activity/pgstat_table.c
@@ -1,9 +1,9 @@
 /* -------------------------------------------------------------------------
  *
- * pgstat_relation.c
- *	  Implementation of relation statistics.
+ * pgstat_table.c
+ *	  Implementation of table statistics.
  *
- * This file contains the implementation of function relation. It is kept
+ * This file contains the implementation of function table. It is kept
  * separate from pgstat.c to enforce the line between the statistics access /
  * storage implementation and the details about individual types of
  * statistics.
@@ -11,7 +11,7 @@
  * Copyright (c) 2001-2023, PostgreSQL Global Development Group
  *
  * IDENTIFICATION
- *	  src/backend/utils/activity/pgstat_relation.c
+ *	  src/backend/utils/activity/pgstat_table.c
  * -------------------------------------------------------------------------
  */
 
@@ -44,74 +44,48 @@ typedef struct TwoPhasePgStatRecord
 } TwoPhasePgStatRecord;
 
 
-static PgStat_TableStatus *pgstat_prep_relation_pending(Oid rel_id, bool isshared);
-static void add_tabstat_xact_level(PgStat_TableStatus *pgstat_info, int nest_level);
-static void ensure_tabstat_xact_level(PgStat_TableStatus *pgstat_info);
+static PgStat_TableStatus *pgstat_prep_table_pending(Oid rel_id, bool isshared);
+static void add_tabstat_xact_level(PgStat_TableStatus *pgstattab_info, int nest_level);
+static void ensure_tabstat_xact_level(PgStat_TableStatus *pgstattab_info);
 static void save_truncdrop_counters(PgStat_TableXactStatus *trans, bool is_drop);
 static void restore_truncdrop_counters(PgStat_TableXactStatus *trans);
 
-
-/*
- * Copy stats between relations. This is used for things like REINDEX
- * CONCURRENTLY.
- */
-void
-pgstat_copy_relation_stats(Relation dst, Relation src)
-{
-	PgStat_StatTabEntry *srcstats;
-	PgStatShared_Relation *dstshstats;
-	PgStat_EntryRef *dst_ref;
-
-	srcstats = pgstat_fetch_stat_tabentry_ext(src->rd_rel->relisshared,
-											  RelationGetRelid(src));
-	if (!srcstats)
-		return;
-
-	dst_ref = pgstat_get_entry_ref_locked(PGSTAT_KIND_RELATION,
-										  dst->rd_rel->relisshared ? InvalidOid : MyDatabaseId,
-										  RelationGetRelid(dst),
-										  false);
-
-	dstshstats = (PgStatShared_Relation *) dst_ref->shared_stats;
-	dstshstats->stats = *srcstats;
-
-	pgstat_unlock_entry(dst_ref);
-}
-
 /*
  * Initialize a relcache entry to count access statistics.  Called whenever a
- * relation is opened.
+ * table is opened.
  *
- * We assume that a relcache entry's pgstat_info field is zeroed by relcache.c
+ * We assume that a relcache entry's pgstattab_info field is zeroed by relcache.c
  * when the relcache entry is made; thereafter it is long-lived data.
  *
  * This does not create a reference to a stats entry in shared memory, nor
  * allocate memory for the pending stats. That happens in
- * pgstat_assoc_relation().
+ * pgstat_assoc_table().
  */
 void
-pgstat_init_relation(Relation rel)
+pgstat_init_table(Relation rel)
 {
 	char		relkind = rel->rd_rel->relkind;
 
+	Assert(relkind != RELKIND_INDEX);
+
 	/*
 	 * We only count stats for relations with storage and partitioned tables
 	 */
 	if (!RELKIND_HAS_STORAGE(relkind) && relkind != RELKIND_PARTITIONED_TABLE)
 	{
 		rel->pgstat_enabled = false;
-		rel->pgstat_info = NULL;
+		rel->pgstattab_info = NULL;
 		return;
 	}
 
 	if (!pgstat_track_counts)
 	{
-		if (rel->pgstat_info)
-			pgstat_unlink_relation(rel);
+		if (rel->pgstattab_info != NULL)
+			pgstat_unlink_table(rel);
 
 		/* We're not counting at all */
 		rel->pgstat_enabled = false;
-		rel->pgstat_info = NULL;
+		rel->pgstattab_info = NULL;
 		return;
 	}
 
@@ -119,89 +93,92 @@ pgstat_init_relation(Relation rel)
 }
 
 /*
- * Prepare for statistics for this relation to be collected.
+ * Prepare for statistics for this table to be collected.
  *
  * This ensures we have a reference to the stats entry before stats can be
- * generated. That is important because a relation drop in another connection
+ * generated. That is important because a table drop in another connection
  * could otherwise lead to the stats entry being dropped, which then later
  * would get recreated when flushing stats.
  *
- * This is separate from pgstat_init_relation() as it is not uncommon for
+ * This is separate from pgstat_init_table() as it is not uncommon for
  * relcache entries to be opened without ever getting stats reported.
  */
 void
-pgstat_assoc_relation(Relation rel)
+pgstat_assoc_table(Relation rel)
 {
 	Assert(rel->pgstat_enabled);
-	Assert(rel->pgstat_info == NULL);
+	Assert(rel->pgstattab_info == NULL);
 
 	/* Else find or make the PgStat_TableStatus entry, and update link */
-	rel->pgstat_info = pgstat_prep_relation_pending(RelationGetRelid(rel),
+	rel->pgstattab_info = pgstat_prep_table_pending(RelationGetRelid(rel),
 													rel->rd_rel->relisshared);
 
 	/* don't allow link a stats to multiple relcache entries */
-	Assert(rel->pgstat_info->relation == NULL);
+	Assert(rel->pgstattab_info->relation == NULL);
 
 	/* mark this relation as the owner */
-	rel->pgstat_info->relation = rel;
+	rel->pgstattab_info->relation = rel;
 }
 
 /*
- * Break the mutual link between a relcache entry and pending stats entry.
+ * Break the mutual link between a relcache entry and pending table stats entry.
  * This must be called whenever one end of the link is removed.
  */
 void
-pgstat_unlink_relation(Relation rel)
+pgstat_unlink_table(Relation rel)
 {
-	/* remove the link to stats info if any */
-	if (rel->pgstat_info == NULL)
+
+	if (rel->pgstattab_info == NULL)
 		return;
 
-	/* link sanity check */
-	Assert(rel->pgstat_info->relation == rel);
-	rel->pgstat_info->relation = NULL;
-	rel->pgstat_info = NULL;
+	/* link sanity check for the table stats */
+	if (rel->pgstattab_info)
+	{
+		Assert(rel->pgstattab_info->relation == rel);
+		rel->pgstattab_info->relation = NULL;
+		rel->pgstattab_info = NULL;
+	}
 }
 
 /*
- * Ensure that stats are dropped if transaction aborts.
+ * Ensure that table stats are dropped if transaction aborts.
  */
 void
-pgstat_create_relation(Relation rel)
+pgstat_create_table(Relation rel)
 {
-	pgstat_create_transactional(PGSTAT_KIND_RELATION,
+	pgstat_create_transactional(PGSTAT_KIND_TABLE,
 								rel->rd_rel->relisshared ? InvalidOid : MyDatabaseId,
 								RelationGetRelid(rel));
 }
 
 /*
- * Ensure that stats are dropped if transaction commits.
+ * Ensure that table stats are dropped if transaction commits.
  */
 void
-pgstat_drop_relation(Relation rel)
+pgstat_drop_table(Relation rel)
 {
 	int			nest_level = GetCurrentTransactionNestLevel();
-	PgStat_TableStatus *pgstat_info;
+	PgStat_TableStatus *pgstattab_info;
 
-	pgstat_drop_transactional(PGSTAT_KIND_RELATION,
+	pgstat_drop_transactional(PGSTAT_KIND_TABLE,
 							  rel->rd_rel->relisshared ? InvalidOid : MyDatabaseId,
 							  RelationGetRelid(rel));
 
-	if (!pgstat_should_count_relation(rel))
+	if (!pgstat_should_count_table(rel))
 		return;
 
 	/*
 	 * Transactionally set counters to 0. That ensures that accesses to
 	 * pg_stat_xact_all_tables inside the transaction show 0.
 	 */
-	pgstat_info = rel->pgstat_info;
-	if (pgstat_info->trans &&
-		pgstat_info->trans->nest_level == nest_level)
+	pgstattab_info = rel->pgstattab_info;
+	if (pgstattab_info->trans &&
+		pgstattab_info->trans->nest_level == nest_level)
 	{
-		save_truncdrop_counters(pgstat_info->trans, true);
-		pgstat_info->trans->tuples_inserted = 0;
-		pgstat_info->trans->tuples_updated = 0;
-		pgstat_info->trans->tuples_deleted = 0;
+		save_truncdrop_counters(pgstattab_info->trans, true);
+		pgstattab_info->trans->tuples_inserted = 0;
+		pgstattab_info->trans->tuples_updated = 0;
+		pgstattab_info->trans->tuples_deleted = 0;
 	}
 }
 
@@ -213,7 +190,7 @@ pgstat_report_vacuum(Oid tableoid, bool shared,
 					 PgStat_Counter livetuples, PgStat_Counter deadtuples)
 {
 	PgStat_EntryRef *entry_ref;
-	PgStatShared_Relation *shtabentry;
+	PgStatShared_Table *shtabentry;
 	PgStat_StatTabEntry *tabentry;
 	Oid			dboid = (shared ? InvalidOid : MyDatabaseId);
 	TimestampTz ts;
@@ -225,10 +202,10 @@ pgstat_report_vacuum(Oid tableoid, bool shared,
 	ts = GetCurrentTimestamp();
 
 	/* block acquiring lock for the same reason as pgstat_report_autovac() */
-	entry_ref = pgstat_get_entry_ref_locked(PGSTAT_KIND_RELATION,
+	entry_ref = pgstat_get_entry_ref_locked(PGSTAT_KIND_TABLE,
 											dboid, tableoid, false);
 
-	shtabentry = (PgStatShared_Relation *) entry_ref->shared_stats;
+	shtabentry = (PgStatShared_Table *) entry_ref->shared_stats;
 	tabentry = &shtabentry->stats;
 
 	tabentry->live_tuples = livetuples;
@@ -280,7 +257,7 @@ pgstat_report_analyze(Relation rel,
 					  bool resetcounter)
 {
 	PgStat_EntryRef *entry_ref;
-	PgStatShared_Relation *shtabentry;
+	PgStatShared_Table *shtabentry;
 	PgStat_StatTabEntry *tabentry;
 	Oid			dboid = (rel->rd_rel->relisshared ? InvalidOid : MyDatabaseId);
 
@@ -299,31 +276,31 @@ pgstat_report_analyze(Relation rel,
 	 *
 	 * Waste no time on partitioned tables, though.
 	 */
-	if (pgstat_should_count_relation(rel) &&
+	if (pgstat_should_count_table(rel) &&
 		rel->rd_rel->relkind != RELKIND_PARTITIONED_TABLE)
 	{
 		PgStat_TableXactStatus *trans;
 
-		for (trans = rel->pgstat_info->trans; trans; trans = trans->upper)
+		for (trans = rel->pgstattab_info->trans; trans; trans = trans->upper)
 		{
 			livetuples -= trans->tuples_inserted - trans->tuples_deleted;
 			deadtuples -= trans->tuples_updated + trans->tuples_deleted;
 		}
 		/* count stuff inserted by already-aborted subxacts, too */
-		deadtuples -= rel->pgstat_info->counts.delta_dead_tuples;
+		deadtuples -= rel->pgstattab_info->counts.delta_dead_tuples;
 		/* Since ANALYZE's counts are estimates, we could have underflowed */
 		livetuples = Max(livetuples, 0);
 		deadtuples = Max(deadtuples, 0);
 	}
 
 	/* block acquiring lock for the same reason as pgstat_report_autovac() */
-	entry_ref = pgstat_get_entry_ref_locked(PGSTAT_KIND_RELATION, dboid,
+	entry_ref = pgstat_get_entry_ref_locked(PGSTAT_KIND_TABLE, dboid,
 											RelationGetRelid(rel),
 											false);
 	/* can't get dropped while accessed */
 	Assert(entry_ref != NULL && entry_ref->shared_stats != NULL);
 
-	shtabentry = (PgStatShared_Relation *) entry_ref->shared_stats;
+	shtabentry = (PgStatShared_Table *) entry_ref->shared_stats;
 	tabentry = &shtabentry->stats;
 
 	tabentry->live_tuples = livetuples;
@@ -360,12 +337,12 @@ pgstat_report_analyze(Relation rel,
 void
 pgstat_count_heap_insert(Relation rel, PgStat_Counter n)
 {
-	if (pgstat_should_count_relation(rel))
+	if (pgstat_should_count_table(rel))
 	{
-		PgStat_TableStatus *pgstat_info = rel->pgstat_info;
+		PgStat_TableStatus *pgstattab_info = rel->pgstattab_info;
 
-		ensure_tabstat_xact_level(pgstat_info);
-		pgstat_info->trans->tuples_inserted += n;
+		ensure_tabstat_xact_level(pgstattab_info);
+		pgstattab_info->trans->tuples_inserted += n;
 	}
 }
 
@@ -377,21 +354,21 @@ pgstat_count_heap_update(Relation rel, bool hot, bool newpage)
 {
 	Assert(!(hot && newpage));
 
-	if (pgstat_should_count_relation(rel))
+	if (pgstat_should_count_table(rel))
 	{
-		PgStat_TableStatus *pgstat_info = rel->pgstat_info;
+		PgStat_TableStatus *pgstattab_info = rel->pgstattab_info;
 
-		ensure_tabstat_xact_level(pgstat_info);
-		pgstat_info->trans->tuples_updated++;
+		ensure_tabstat_xact_level(pgstattab_info);
+		pgstattab_info->trans->tuples_updated++;
 
 		/*
 		 * tuples_hot_updated and tuples_newpage_updated counters are
 		 * nontransactional, so just advance them
 		 */
 		if (hot)
-			pgstat_info->counts.tuples_hot_updated++;
+			pgstattab_info->counts.tuples_hot_updated++;
 		else if (newpage)
-			pgstat_info->counts.tuples_newpage_updated++;
+			pgstattab_info->counts.tuples_newpage_updated++;
 	}
 }
 
@@ -401,12 +378,12 @@ pgstat_count_heap_update(Relation rel, bool hot, bool newpage)
 void
 pgstat_count_heap_delete(Relation rel)
 {
-	if (pgstat_should_count_relation(rel))
+	if (pgstat_should_count_table(rel))
 	{
-		PgStat_TableStatus *pgstat_info = rel->pgstat_info;
+		PgStat_TableStatus *pgstattab_info = rel->pgstattab_info;
 
-		ensure_tabstat_xact_level(pgstat_info);
-		pgstat_info->trans->tuples_deleted++;
+		ensure_tabstat_xact_level(pgstattab_info);
+		pgstattab_info->trans->tuples_deleted++;
 	}
 }
 
@@ -416,15 +393,15 @@ pgstat_count_heap_delete(Relation rel)
 void
 pgstat_count_truncate(Relation rel)
 {
-	if (pgstat_should_count_relation(rel))
+	if (pgstat_should_count_table(rel))
 	{
-		PgStat_TableStatus *pgstat_info = rel->pgstat_info;
+		PgStat_TableStatus *pgstattab_info = rel->pgstattab_info;
 
-		ensure_tabstat_xact_level(pgstat_info);
-		save_truncdrop_counters(pgstat_info->trans, false);
-		pgstat_info->trans->tuples_inserted = 0;
-		pgstat_info->trans->tuples_updated = 0;
-		pgstat_info->trans->tuples_deleted = 0;
+		ensure_tabstat_xact_level(pgstattab_info);
+		save_truncdrop_counters(pgstattab_info->trans, false);
+		pgstattab_info->trans->tuples_inserted = 0;
+		pgstattab_info->trans->tuples_updated = 0;
+		pgstattab_info->trans->tuples_deleted = 0;
 	}
 }
 
@@ -439,11 +416,11 @@ pgstat_count_truncate(Relation rel)
 void
 pgstat_update_heap_dead_tuples(Relation rel, int delta)
 {
-	if (pgstat_should_count_relation(rel))
+	if (pgstat_should_count_table(rel))
 	{
-		PgStat_TableStatus *pgstat_info = rel->pgstat_info;
+		PgStat_TableStatus *pgstattab_info = rel->pgstattab_info;
 
-		pgstat_info->counts.delta_dead_tuples -= delta;
+		pgstattab_info->counts.delta_dead_tuples -= delta;
 	}
 }
 
@@ -469,7 +446,7 @@ pgstat_fetch_stat_tabentry_ext(bool shared, Oid reloid)
 	Oid			dboid = (shared ? InvalidOid : MyDatabaseId);
 
 	return (PgStat_StatTabEntry *)
-		pgstat_fetch_entry(PGSTAT_KIND_RELATION, dboid, reloid);
+		pgstat_fetch_entry(PGSTAT_KIND_TABLE, dboid, reloid);
 }
 
 /*
@@ -492,10 +469,10 @@ find_tabstat_entry(Oid rel_id)
 	PgStat_TableStatus *tabentry = NULL;
 	PgStat_TableStatus *tablestatus = NULL;
 
-	entry_ref = pgstat_fetch_pending_entry(PGSTAT_KIND_RELATION, MyDatabaseId, rel_id);
+	entry_ref = pgstat_fetch_pending_entry(PGSTAT_KIND_TABLE, MyDatabaseId, rel_id);
 	if (!entry_ref)
 	{
-		entry_ref = pgstat_fetch_pending_entry(PGSTAT_KIND_RELATION, InvalidOid, rel_id);
+		entry_ref = pgstat_fetch_pending_entry(PGSTAT_KIND_TABLE, InvalidOid, rel_id);
 		if (!entry_ref)
 			return tablestatus;
 	}
@@ -527,7 +504,7 @@ find_tabstat_entry(Oid rel_id)
 }
 
 /*
- * Perform relation stats specific end-of-transaction work. Helper for
+ * Perform table stats specific end-of-transaction work. Helper for
  * AtEOXact_PgStat.
  *
  * Transfer transactional insert/update counts into the base tabstat entries.
@@ -535,7 +512,7 @@ find_tabstat_entry(Oid rel_id)
  * TopTransactionContext and will go away anyway.
  */
 void
-AtEOXact_PgStat_Relations(PgStat_SubXactStatus *xact_state, bool isCommit)
+AtEOXact_PgStat_Tables(PgStat_SubXactStatus *xact_state, bool isCommit)
 {
 	PgStat_TableXactStatus *trans;
 
@@ -586,14 +563,14 @@ AtEOXact_PgStat_Relations(PgStat_SubXactStatus *xact_state, bool isCommit)
 }
 
 /*
- * Perform relation stats specific end-of-sub-transaction work. Helper for
+ * Perform table stats specific end-of-sub-transaction work. Helper for
  * AtEOSubXact_PgStat.
  *
  * Transfer transactional insert/update counts into the next higher
  * subtransaction state.
  */
 void
-AtEOSubXact_PgStat_Relations(PgStat_SubXactStatus *xact_state, bool isCommit, int nestDepth)
+AtEOSubXact_PgStat_Tables(PgStat_SubXactStatus *xact_state, bool isCommit, int nestDepth)
 {
 	PgStat_TableXactStatus *trans;
 	PgStat_TableXactStatus *next_trans;
@@ -670,11 +647,11 @@ AtEOSubXact_PgStat_Relations(PgStat_SubXactStatus *xact_state, bool isCommit, in
 }
 
 /*
- * Generate 2PC records for all the pending transaction-dependent relation
+ * Generate 2PC records for all the pending transaction-dependent table
  * stats.
  */
 void
-AtPrepare_PgStat_Relations(PgStat_SubXactStatus *xact_state)
+AtPrepare_PgStat_Tables(PgStat_SubXactStatus *xact_state)
 {
 	PgStat_TableXactStatus *trans;
 
@@ -709,10 +686,10 @@ AtPrepare_PgStat_Relations(PgStat_SubXactStatus *xact_state)
  * reported to the stats system immediately, while the effects on live and
  * dead tuple counts are preserved in the 2PC state file.
  *
- * Note: AtEOXact_PgStat_Relations is not called during PREPARE.
+ * Note: AtEOXact_PgStat_Tables is not called during PREPARE.
  */
 void
-PostPrepare_PgStat_Relations(PgStat_SubXactStatus *xact_state)
+PostPrepare_PgStat_Tables(PgStat_SubXactStatus *xact_state)
 {
 	PgStat_TableXactStatus *trans;
 
@@ -735,27 +712,27 @@ pgstat_twophase_postcommit(TransactionId xid, uint16 info,
 						   void *recdata, uint32 len)
 {
 	TwoPhasePgStatRecord *rec = (TwoPhasePgStatRecord *) recdata;
-	PgStat_TableStatus *pgstat_info;
+	PgStat_TableStatus *pgstattab_info;
 
-	/* Find or create a tabstat entry for the rel */
-	pgstat_info = pgstat_prep_relation_pending(rec->id, rec->shared);
+	/* Find or create a tabstat entry for the table */
+	pgstattab_info = pgstat_prep_table_pending(rec->id, rec->shared);
 
 	/* Same math as in AtEOXact_PgStat, commit case */
-	pgstat_info->counts.tuples_inserted += rec->tuples_inserted;
-	pgstat_info->counts.tuples_updated += rec->tuples_updated;
-	pgstat_info->counts.tuples_deleted += rec->tuples_deleted;
-	pgstat_info->counts.truncdropped = rec->truncdropped;
+	pgstattab_info->counts.tuples_inserted += rec->tuples_inserted;
+	pgstattab_info->counts.tuples_updated += rec->tuples_updated;
+	pgstattab_info->counts.tuples_deleted += rec->tuples_deleted;
+	pgstattab_info->counts.truncdropped = rec->truncdropped;
 	if (rec->truncdropped)
 	{
 		/* forget live/dead stats seen by backend thus far */
-		pgstat_info->counts.delta_live_tuples = 0;
-		pgstat_info->counts.delta_dead_tuples = 0;
+		pgstattab_info->counts.delta_live_tuples = 0;
+		pgstattab_info->counts.delta_dead_tuples = 0;
 	}
-	pgstat_info->counts.delta_live_tuples +=
+	pgstattab_info->counts.delta_live_tuples +=
 		rec->tuples_inserted - rec->tuples_deleted;
-	pgstat_info->counts.delta_dead_tuples +=
+	pgstattab_info->counts.delta_dead_tuples +=
 		rec->tuples_updated + rec->tuples_deleted;
-	pgstat_info->counts.changed_tuples +=
+	pgstattab_info->counts.changed_tuples +=
 		rec->tuples_inserted + rec->tuples_updated +
 		rec->tuples_deleted;
 }
@@ -771,10 +748,10 @@ pgstat_twophase_postabort(TransactionId xid, uint16 info,
 						  void *recdata, uint32 len)
 {
 	TwoPhasePgStatRecord *rec = (TwoPhasePgStatRecord *) recdata;
-	PgStat_TableStatus *pgstat_info;
+	PgStat_TableStatus *pgstattab_info;
 
-	/* Find or create a tabstat entry for the rel */
-	pgstat_info = pgstat_prep_relation_pending(rec->id, rec->shared);
+	/* Find or create a tabstat entry for the table */
+	pgstattab_info = pgstat_prep_table_pending(rec->id, rec->shared);
 
 	/* Same math as in AtEOXact_PgStat, abort case */
 	if (rec->truncdropped)
@@ -783,10 +760,10 @@ pgstat_twophase_postabort(TransactionId xid, uint16 info,
 		rec->tuples_updated = rec->updated_pre_truncdrop;
 		rec->tuples_deleted = rec->deleted_pre_truncdrop;
 	}
-	pgstat_info->counts.tuples_inserted += rec->tuples_inserted;
-	pgstat_info->counts.tuples_updated += rec->tuples_updated;
-	pgstat_info->counts.tuples_deleted += rec->tuples_deleted;
-	pgstat_info->counts.delta_dead_tuples +=
+	pgstattab_info->counts.tuples_inserted += rec->tuples_inserted;
+	pgstattab_info->counts.tuples_updated += rec->tuples_updated;
+	pgstattab_info->counts.tuples_deleted += rec->tuples_deleted;
+	pgstattab_info->counts.delta_dead_tuples +=
 		rec->tuples_inserted + rec->tuples_updated;
 }
 
@@ -800,22 +777,21 @@ pgstat_twophase_postabort(TransactionId xid, uint16 info,
  * entry when successfully flushing.
  */
 bool
-pgstat_relation_flush_cb(PgStat_EntryRef *entry_ref, bool nowait)
+pgstat_table_flush_cb(PgStat_EntryRef *entry_ref, bool nowait)
 {
 	static const PgStat_TableCounts all_zeroes;
 	Oid			dboid;
 	PgStat_TableStatus *lstats; /* pending stats entry  */
-	PgStatShared_Relation *shtabstats;
+	PgStatShared_Table *shtabstats;
 	PgStat_StatTabEntry *tabentry;	/* table entry of shared stats */
 	PgStat_StatDBEntry *dbentry;	/* pending database entry */
 
 	dboid = entry_ref->shared_entry->key.dboid;
 	lstats = (PgStat_TableStatus *) entry_ref->pending;
-	shtabstats = (PgStatShared_Relation *) entry_ref->shared_stats;
+	shtabstats = (PgStatShared_Table *) entry_ref->shared_stats;
 
 	/*
-	 * Ignore entries that didn't accumulate any actual counts, such as
-	 * indexes that were opened by the planner but not used.
+	 * Ignore entries that didn't accumulate any actual counts.
 	 */
 	if (memcmp(&lstats->counts, &all_zeroes,
 			   sizeof(PgStat_TableCounts)) == 0)
@@ -883,12 +859,12 @@ pgstat_relation_flush_cb(PgStat_EntryRef *entry_ref, bool nowait)
 }
 
 void
-pgstat_relation_delete_pending_cb(PgStat_EntryRef *entry_ref)
+pgstat_table_delete_pending_cb(PgStat_EntryRef *entry_ref)
 {
 	PgStat_TableStatus *pending = (PgStat_TableStatus *) entry_ref->pending;
 
 	if (pending->relation)
-		pgstat_unlink_relation(pending->relation);
+		pgstat_unlink_table(pending->relation);
 }
 
 /*
@@ -896,12 +872,12 @@ pgstat_relation_delete_pending_cb(PgStat_EntryRef *entry_ref)
  * initialized if not exists.
  */
 static PgStat_TableStatus *
-pgstat_prep_relation_pending(Oid rel_id, bool isshared)
+pgstat_prep_table_pending(Oid rel_id, bool isshared)
 {
 	PgStat_EntryRef *entry_ref;
 	PgStat_TableStatus *pending;
 
-	entry_ref = pgstat_prep_pending_entry(PGSTAT_KIND_RELATION,
+	entry_ref = pgstat_prep_pending_entry(PGSTAT_KIND_TABLE,
 										  isshared ? InvalidOid : MyDatabaseId,
 										  rel_id, NULL);
 	pending = entry_ref->pending;
@@ -915,7 +891,7 @@ pgstat_prep_relation_pending(Oid rel_id, bool isshared)
  * add a new (sub)transaction state record
  */
 static void
-add_tabstat_xact_level(PgStat_TableStatus *pgstat_info, int nest_level)
+add_tabstat_xact_level(PgStat_TableStatus *pgstattab_info, int nest_level)
 {
 	PgStat_SubXactStatus *xact_state;
 	PgStat_TableXactStatus *trans;
@@ -931,24 +907,24 @@ add_tabstat_xact_level(PgStat_TableStatus *pgstat_info, int nest_level)
 		MemoryContextAllocZero(TopTransactionContext,
 							   sizeof(PgStat_TableXactStatus));
 	trans->nest_level = nest_level;
-	trans->upper = pgstat_info->trans;
-	trans->parent = pgstat_info;
+	trans->upper = pgstattab_info->trans;
+	trans->parent = pgstattab_info;
 	trans->next = xact_state->first;
 	xact_state->first = trans;
-	pgstat_info->trans = trans;
+	pgstattab_info->trans = trans;
 }
 
 /*
  * Add a new (sub)transaction record if needed.
  */
 static void
-ensure_tabstat_xact_level(PgStat_TableStatus *pgstat_info)
+ensure_tabstat_xact_level(PgStat_TableStatus *pgstattab_info)
 {
 	int			nest_level = GetCurrentTransactionNestLevel();
 
-	if (pgstat_info->trans == NULL ||
-		pgstat_info->trans->nest_level != nest_level)
-		add_tabstat_xact_level(pgstat_info, nest_level);
+	if (pgstattab_info->trans == NULL ||
+		pgstattab_info->trans->nest_level != nest_level)
+		add_tabstat_xact_level(pgstattab_info, nest_level);
 }
 
 /*
diff --git a/src/backend/utils/activity/pgstat_xact.c b/src/backend/utils/activity/pgstat_xact.c
index 369239d501..cc87f3d015 100644
--- a/src/backend/utils/activity/pgstat_xact.c
+++ b/src/backend/utils/activity/pgstat_xact.c
@@ -51,7 +51,7 @@ AtEOXact_PgStat(bool isCommit, bool parallel)
 		Assert(xact_state->nest_level == 1);
 		Assert(xact_state->prev == NULL);
 
-		AtEOXact_PgStat_Relations(xact_state, isCommit);
+		AtEOXact_PgStat_Tables(xact_state, isCommit);
 		AtEOXact_PgStat_DroppedStats(xact_state, isCommit);
 	}
 	pgStatXactStack = NULL;
@@ -122,7 +122,7 @@ AtEOSubXact_PgStat(bool isCommit, int nestDepth)
 		/* delink xact_state from stack immediately to simplify reuse case */
 		pgStatXactStack = xact_state->prev;
 
-		AtEOSubXact_PgStat_Relations(xact_state, isCommit, nestDepth);
+		AtEOSubXact_PgStat_Tables(xact_state, isCommit, nestDepth);
 		AtEOSubXact_PgStat_DroppedStats(xact_state, isCommit, nestDepth);
 
 		pfree(xact_state);
@@ -197,7 +197,7 @@ AtPrepare_PgStat(void)
 		Assert(xact_state->nest_level == 1);
 		Assert(xact_state->prev == NULL);
 
-		AtPrepare_PgStat_Relations(xact_state);
+		AtPrepare_PgStat_Tables(xact_state);
 	}
 }
 
@@ -221,7 +221,7 @@ PostPrepare_PgStat(void)
 		Assert(xact_state->nest_level == 1);
 		Assert(xact_state->prev == NULL);
 
-		PostPrepare_PgStat_Relations(xact_state);
+		PostPrepare_PgStat_Tables(xact_state);
 	}
 	pgStatXactStack = NULL;
 
diff --git a/src/backend/utils/adt/pgstatfuncs.c b/src/backend/utils/adt/pgstatfuncs.c
index 3a9f9bc4fe..62a49dd686 100644
--- a/src/backend/utils/adt/pgstatfuncs.c
+++ b/src/backend/utils/adt/pgstatfuncs.c
@@ -38,9 +38,9 @@
 
 #define HAS_PGSTAT_PERMISSIONS(role)	 (has_privs_of_role(GetUserId(), ROLE_PG_READ_ALL_STATS) || has_privs_of_role(GetUserId(), role))
 
-#define PG_STAT_GET_RELENTRY_INT64(stat)						\
+#define PG_STAT_GET_TABENTRY_INT64(stat)						\
 Datum															\
-CppConcat(pg_stat_get_,stat)(PG_FUNCTION_ARGS)					\
+CppConcat(pg_stat_get_tab_,stat)(PG_FUNCTION_ARGS)				\
 {																\
 	Oid			relid = PG_GETARG_OID(0);						\
 	int64		result;											\
@@ -54,63 +54,95 @@ CppConcat(pg_stat_get_,stat)(PG_FUNCTION_ARGS)					\
 	PG_RETURN_INT64(result);									\
 }
 
-/* pg_stat_get_analyze_count */
-PG_STAT_GET_RELENTRY_INT64(analyze_count)
+/* pg_stat_get_tab_analyze_count */
+PG_STAT_GET_TABENTRY_INT64(analyze_count)
 
-/* pg_stat_get_autoanalyze_count */
-PG_STAT_GET_RELENTRY_INT64(autoanalyze_count)
+/* pg_stat_get_tab_autoanalyze_count */
+PG_STAT_GET_TABENTRY_INT64(autoanalyze_count)
 
-/* pg_stat_get_autovacuum_count */
-PG_STAT_GET_RELENTRY_INT64(autovacuum_count)
+/* pg_stat_get_tab_autovacuum_count */
+PG_STAT_GET_TABENTRY_INT64(autovacuum_count)
 
-/* pg_stat_get_blocks_fetched */
-PG_STAT_GET_RELENTRY_INT64(blocks_fetched)
+/* pg_stat_get_tab_blocks_fetched */
+PG_STAT_GET_TABENTRY_INT64(blocks_fetched)
 
-/* pg_stat_get_blocks_hit */
-PG_STAT_GET_RELENTRY_INT64(blocks_hit)
+/* pg_stat_get_tab_blocks_hit */
+PG_STAT_GET_TABENTRY_INT64(blocks_hit)
 
-/* pg_stat_get_dead_tuples */
-PG_STAT_GET_RELENTRY_INT64(dead_tuples)
+/* pg_stat_get_tab_dead_tuples */
+PG_STAT_GET_TABENTRY_INT64(dead_tuples)
 
-/* pg_stat_get_ins_since_vacuum */
-PG_STAT_GET_RELENTRY_INT64(ins_since_vacuum)
+/* pg_stat_get_tab_ins_since_vacuum */
+PG_STAT_GET_TABENTRY_INT64(ins_since_vacuum)
 
-/* pg_stat_get_live_tuples */
-PG_STAT_GET_RELENTRY_INT64(live_tuples)
+/* pg_stat_get_tab_live_tuples */
+PG_STAT_GET_TABENTRY_INT64(live_tuples)
 
-/* pg_stat_get_mod_since_analyze */
-PG_STAT_GET_RELENTRY_INT64(mod_since_analyze)
+/* pg_stat_get_tab_mod_since_analyze */
+PG_STAT_GET_TABENTRY_INT64(mod_since_analyze)
 
-/* pg_stat_get_numscans */
-PG_STAT_GET_RELENTRY_INT64(numscans)
+/* pg_stat_get_tab_numscans */
+PG_STAT_GET_TABENTRY_INT64(numscans)
 
-/* pg_stat_get_tuples_deleted */
-PG_STAT_GET_RELENTRY_INT64(tuples_deleted)
+/* pg_stat_get_tab_tuples_deleted */
+PG_STAT_GET_TABENTRY_INT64(tuples_deleted)
 
-/* pg_stat_get_tuples_fetched */
-PG_STAT_GET_RELENTRY_INT64(tuples_fetched)
+/* pg_stat_get_tab_tuples_fetched */
+PG_STAT_GET_TABENTRY_INT64(tuples_fetched)
 
-/* pg_stat_get_tuples_hot_updated */
-PG_STAT_GET_RELENTRY_INT64(tuples_hot_updated)
+/* pg_stat_get_tab_tuples_hot_updated */
+PG_STAT_GET_TABENTRY_INT64(tuples_hot_updated)
 
-/* pg_stat_get_tuples_newpage_updated */
-PG_STAT_GET_RELENTRY_INT64(tuples_newpage_updated)
+/* pg_stat_get_tab_tuples_newpage_updated */
+PG_STAT_GET_TABENTRY_INT64(tuples_newpage_updated)
 
-/* pg_stat_get_tuples_inserted */
-PG_STAT_GET_RELENTRY_INT64(tuples_inserted)
+/* pg_stat_get_tab_tuples_inserted */
+PG_STAT_GET_TABENTRY_INT64(tuples_inserted)
 
-/* pg_stat_get_tuples_returned */
-PG_STAT_GET_RELENTRY_INT64(tuples_returned)
+/* pg_stat_get_tab_tuples_returned */
+PG_STAT_GET_TABENTRY_INT64(tuples_returned)
 
-/* pg_stat_get_tuples_updated */
-PG_STAT_GET_RELENTRY_INT64(tuples_updated)
+/* pg_stat_get_tab_tuples_updated */
+PG_STAT_GET_TABENTRY_INT64(tuples_updated)
 
-/* pg_stat_get_vacuum_count */
-PG_STAT_GET_RELENTRY_INT64(vacuum_count)
+/* pg_stat_get_tab_vacuum_count */
+PG_STAT_GET_TABENTRY_INT64(vacuum_count)
 
-#define PG_STAT_GET_RELENTRY_TIMESTAMPTZ(stat)					\
+
+#define PG_STAT_GET_INDENTRY_INT64(stat)						\
 Datum															\
-CppConcat(pg_stat_get_,stat)(PG_FUNCTION_ARGS)					\
+CppConcat(pg_stat_get_ind_,stat)(PG_FUNCTION_ARGS)				\
+{																\
+	Oid			relid = PG_GETARG_OID(0);						\
+	int64		result;											\
+	PgStat_StatIndEntry *indentry;								\
+																\
+	if ((indentry = pgstat_fetch_stat_indentry(relid)) == NULL)	\
+		result = 0;												\
+	else														\
+		result = (int64) (indentry->stat);						\
+																\
+	PG_RETURN_INT64(result);									\
+}
+
+/* pg_stat_get_ind_numscans */
+PG_STAT_GET_INDENTRY_INT64(numscans)
+
+/* pg_stat_get_ind_tuples_returned */
+PG_STAT_GET_INDENTRY_INT64(tuples_returned)
+
+/* pg_stat_get_ind_tuples_fetched */
+PG_STAT_GET_INDENTRY_INT64(tuples_fetched)
+
+/* pg_stat_get_ind_blocks_fetched */
+PG_STAT_GET_INDENTRY_INT64(blocks_fetched)
+
+/* pg_stat_get_ind_blocks_hit */
+PG_STAT_GET_INDENTRY_INT64(blocks_hit)
+
+#define PG_STAT_GET_TABENTRY_TIMESTAMPTZ(stat)					\
+Datum															\
+CppConcat(pg_stat_get_tab_,stat)(PG_FUNCTION_ARGS)				\
 {																\
 	Oid			relid = PG_GETARG_OID(0);						\
 	TimestampTz result;											\
@@ -127,20 +159,42 @@ CppConcat(pg_stat_get_,stat)(PG_FUNCTION_ARGS)					\
 		PG_RETURN_TIMESTAMPTZ(result);							\
 }
 
-/* pg_stat_get_last_analyze_time */
-PG_STAT_GET_RELENTRY_TIMESTAMPTZ(last_analyze_time)
+/* pg_stat_get_tab_last_analyze_time */
+PG_STAT_GET_TABENTRY_TIMESTAMPTZ(last_analyze_time)
 
-/* pg_stat_get_last_autoanalyze_time */
-PG_STAT_GET_RELENTRY_TIMESTAMPTZ(last_autoanalyze_time)
+/* pg_stat_get_tab_last_autoanalyze_time */
+PG_STAT_GET_TABENTRY_TIMESTAMPTZ(last_autoanalyze_time)
 
-/* pg_stat_get_last_autovacuum_time */
-PG_STAT_GET_RELENTRY_TIMESTAMPTZ(last_autovacuum_time)
+/* pg_stat_get_tab_last_autovacuum_time */
+PG_STAT_GET_TABENTRY_TIMESTAMPTZ(last_autovacuum_time)
 
-/* pg_stat_get_last_vacuum_time */
-PG_STAT_GET_RELENTRY_TIMESTAMPTZ(last_vacuum_time)
+/* pg_stat_get_tab_last_vacuum_time */
+PG_STAT_GET_TABENTRY_TIMESTAMPTZ(last_vacuum_time)
 
-/* pg_stat_get_lastscan */
-PG_STAT_GET_RELENTRY_TIMESTAMPTZ(lastscan)
+/* pg_stat_get_tab_lastscan */
+PG_STAT_GET_TABENTRY_TIMESTAMPTZ(lastscan)
+
+#define PG_STAT_GET_INDENTRY_TIMESTAMPTZ(stat)					\
+Datum															\
+CppConcat(pg_stat_get_ind_,stat)(PG_FUNCTION_ARGS)				\
+{																\
+	Oid			relid = PG_GETARG_OID(0);						\
+	TimestampTz result;											\
+	PgStat_StatIndEntry *indentry;								\
+																\
+	if ((indentry = pgstat_fetch_stat_indentry(relid)) == NULL)	\
+		result = 0;												\
+	else														\
+		result = indentry->stat;								\
+																\
+	if (result == 0)											\
+		PG_RETURN_NULL();										\
+	else														\
+		PG_RETURN_TIMESTAMPTZ(result);							\
+}
+
+/* pg_stat_get_ind_lastscan */
+PG_STAT_GET_INDENTRY_TIMESTAMPTZ(lastscan)
 
 Datum
 pg_stat_get_function_calls(PG_FUNCTION_ARGS)
@@ -1557,9 +1611,9 @@ pg_stat_get_slru(PG_FUNCTION_ARGS)
 	return (Datum) 0;
 }
 
-#define PG_STAT_GET_XACT_RELENTRY_INT64(stat)			\
+#define PG_STAT_GET_XACT_TABENTRY_INT64(stat)			\
 Datum													\
-CppConcat(pg_stat_get_xact_,stat)(PG_FUNCTION_ARGS)		\
+CppConcat(pg_stat_get_tab_xact_,stat)(PG_FUNCTION_ARGS)	\
 {														\
 	Oid         relid = PG_GETARG_OID(0);				\
 	int64       result;									\
@@ -1573,35 +1627,57 @@ CppConcat(pg_stat_get_xact_,stat)(PG_FUNCTION_ARGS)		\
 	PG_RETURN_INT64(result);							\
 }
 
-/* pg_stat_get_xact_numscans */
-PG_STAT_GET_XACT_RELENTRY_INT64(numscans)
+/* pg_stat_get_tab_xact_numscans */
+PG_STAT_GET_XACT_TABENTRY_INT64(numscans)
+
+/* pg_stat_get_tab_xact_tuples_returned */
+PG_STAT_GET_XACT_TABENTRY_INT64(tuples_returned)
+
+/* pg_stat_get_tab_xact_tuples_fetched */
+PG_STAT_GET_XACT_TABENTRY_INT64(tuples_fetched)
 
-/* pg_stat_get_xact_tuples_returned */
-PG_STAT_GET_XACT_RELENTRY_INT64(tuples_returned)
+/* pg_stat_get_tab_xact_tuples_hot_updated */
+PG_STAT_GET_XACT_TABENTRY_INT64(tuples_hot_updated)
 
-/* pg_stat_get_xact_tuples_fetched */
-PG_STAT_GET_XACT_RELENTRY_INT64(tuples_fetched)
+/* pg_stat_get_tab_xact_tuples_newpage_updated */
+PG_STAT_GET_XACT_TABENTRY_INT64(tuples_newpage_updated)
 
-/* pg_stat_get_xact_tuples_hot_updated */
-PG_STAT_GET_XACT_RELENTRY_INT64(tuples_hot_updated)
+/* pg_stat_get_tab_xact_blocks_fetched */
+PG_STAT_GET_XACT_TABENTRY_INT64(blocks_fetched)
 
-/* pg_stat_get_xact_tuples_newpage_updated */
-PG_STAT_GET_XACT_RELENTRY_INT64(tuples_newpage_updated)
+/* pg_stat_get_tab_xact_blocks_hit */
+PG_STAT_GET_XACT_TABENTRY_INT64(blocks_hit)
 
-/* pg_stat_get_xact_blocks_fetched */
-PG_STAT_GET_XACT_RELENTRY_INT64(blocks_fetched)
+/* pg_stat_get_tab_xact_tuples_inserted */
+PG_STAT_GET_XACT_TABENTRY_INT64(tuples_inserted)
 
-/* pg_stat_get_xact_blocks_hit */
-PG_STAT_GET_XACT_RELENTRY_INT64(blocks_hit)
+/* pg_stat_get_tab_xact_tuples_updated */
+PG_STAT_GET_XACT_TABENTRY_INT64(tuples_updated)
 
-/* pg_stat_get_xact_tuples_inserted */
-PG_STAT_GET_XACT_RELENTRY_INT64(tuples_inserted)
+/* pg_stat_get_tab_xact_tuples_deleted */
+PG_STAT_GET_XACT_TABENTRY_INT64(tuples_deleted)
+
+#define PG_STAT_GET_XACT_INDENTRY_INT64(stat)			\
+Datum													\
+CppConcat(pg_stat_get_ind_xact_,stat)(PG_FUNCTION_ARGS)	\
+{														\
+	Oid         relid = PG_GETARG_OID(0);				\
+	int64       result;									\
+	PgStat_IndexStatus *indentry;						\
+														\
+	if ((indentry = find_indstat_entry(relid)) == NULL)	\
+		result = 0;										\
+	else												\
+		result = (int64) (indentry->counts.stat);		\
+														\
+	PG_RETURN_INT64(result);							\
+}
 
-/* pg_stat_get_xact_tuples_updated */
-PG_STAT_GET_XACT_RELENTRY_INT64(tuples_updated)
+/* pg_stat_get_ind_xact_numscans */
+PG_STAT_GET_XACT_INDENTRY_INT64(numscans)
 
-/* pg_stat_get_xact_tuples_deleted */
-PG_STAT_GET_XACT_RELENTRY_INT64(tuples_deleted)
+/* pg_stat_get_ind_xact_tuples_fetched */
+PG_STAT_GET_XACT_INDENTRY_INT64(tuples_fetched)
 
 Datum
 pg_stat_get_xact_function_calls(PG_FUNCTION_ARGS)
@@ -1733,7 +1809,8 @@ pg_stat_reset_single_table_counters(PG_FUNCTION_ARGS)
 	Oid			taboid = PG_GETARG_OID(0);
 	Oid			dboid = (IsSharedRelation(taboid) ? InvalidOid : MyDatabaseId);
 
-	pgstat_reset(PGSTAT_KIND_RELATION, dboid, taboid);
+	pgstat_reset(PGSTAT_KIND_TABLE, dboid, taboid);
+	pgstat_reset(PGSTAT_KIND_INDEX, dboid, taboid);
 
 	PG_RETURN_VOID();
 }
diff --git a/src/backend/utils/cache/relcache.c b/src/backend/utils/cache/relcache.c
index b3faccbefe..e35fb1fe3c 100644
--- a/src/backend/utils/cache/relcache.c
+++ b/src/backend/utils/cache/relcache.c
@@ -2445,7 +2445,10 @@ RelationDestroyRelation(Relation relation, bool remember_tupdesc)
 	RelationCloseSmgr(relation);
 
 	/* break mutual link with stats entry */
-	pgstat_unlink_relation(relation);
+	if (relation->rd_rel->relkind == RELKIND_INDEX)
+		pgstat_unlink_index(relation);
+	else
+		pgstat_unlink_table(relation);
 
 	/*
 	 * Free all the subsidiary data structures of the relcache entry, then the
@@ -2762,8 +2765,9 @@ RelationClearRelation(Relation relation, bool rebuild)
 			SWAPFIELD(RowSecurityDesc *, rd_rsdesc);
 		/* toast OID override must be preserved */
 		SWAPFIELD(Oid, rd_toastoid);
-		/* pgstat_info / enabled must be preserved */
-		SWAPFIELD(struct PgStat_TableStatus *, pgstat_info);
+		/* pgstattab_info / pgstatind_info / enabled must be preserved */
+		SWAPFIELD(struct PgStat_TableStatus *, pgstattab_info);
+		SWAPFIELD(struct PgStat_IndexStatus *, pgstatind_info);
 		SWAPFIELD(bool, pgstat_enabled);
 		/* preserve old partition key if we have one */
 		if (keep_partkey)
@@ -6384,7 +6388,8 @@ load_relcache_init_file(bool shared)
 		rel->rd_firstRelfilelocatorSubid = InvalidSubTransactionId;
 		rel->rd_droppedSubid = InvalidSubTransactionId;
 		rel->rd_amcache = NULL;
-		rel->pgstat_info = NULL;
+		rel->pgstattab_info = NULL;
+		rel->pgstatind_info = NULL;
 
 		/*
 		 * Recompute lock and physical addressing info.  This is needed in
diff --git a/src/include/catalog/pg_proc.dat b/src/include/catalog/pg_proc.dat
index bd0b8873d3..03b2f67e44 100644
--- a/src/include/catalog/pg_proc.dat
+++ b/src/include/catalog/pg_proc.dat
@@ -5324,101 +5324,125 @@
   proargnames => '{mcv_list,index,values,nulls,frequency,base_frequency}',
   prosrc => 'pg_stats_ext_mcvlist_items' },
 
-{ oid => '1928', descr => 'statistics: number of scans done for table/index',
-  proname => 'pg_stat_get_numscans', provolatile => 's', proparallel => 'r',
+{ oid => '1928', descr => 'statistics: number of scans done for table',
+  proname => 'pg_stat_get_tab_numscans', provolatile => 's', proparallel => 'r',
   prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_numscans' },
-{ oid => '6310', descr => 'statistics: time of the last scan for table/index',
-  proname => 'pg_stat_get_lastscan', provolatile => 's', proparallel => 'r',
+  prosrc => 'pg_stat_get_tab_numscans' },
+{ oid => '8808', descr => 'statistics: number of scans done for index',
+  proname => 'pg_stat_get_ind_numscans', provolatile => 's', proparallel => 'r',
+  prorettype => 'int8', proargtypes => 'oid',
+  prosrc => 'pg_stat_get_ind_numscans' },
+{ oid => '6310', descr => 'statistics: time of the last scan for table',
+  proname => 'pg_stat_get_tab_lastscan', provolatile => 's', proparallel => 'r',
+  prorettype => 'timestamptz', proargtypes => 'oid',
+  prosrc => 'pg_stat_get_tab_lastscan' },
+{ oid => '9623', descr => 'statistics: time of the last scan for index',
+  proname => 'pg_stat_get_ind_lastscan', provolatile => 's', proparallel => 'r',
   prorettype => 'timestamptz', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_lastscan' },
+  prosrc => 'pg_stat_get_ind_lastscan' },
 { oid => '1929', descr => 'statistics: number of tuples read by seqscan',
-  proname => 'pg_stat_get_tuples_returned', provolatile => 's',
+  proname => 'pg_stat_get_tab_tuples_returned', provolatile => 's',
+  proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
+  prosrc => 'pg_stat_get_tab_tuples_returned' },
+{ oid => '9362', descr => 'statistics: number of tuples read by index scan',
+  proname => 'pg_stat_get_ind_tuples_returned', provolatile => 's',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_tuples_returned' },
+  prosrc => 'pg_stat_get_ind_tuples_returned' },
 { oid => '1930', descr => 'statistics: number of tuples fetched by idxscan',
-  proname => 'pg_stat_get_tuples_fetched', provolatile => 's',
+  proname => 'pg_stat_get_tab_tuples_fetched', provolatile => 's',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_tuples_fetched' },
+  prosrc => 'pg_stat_get_tab_tuples_fetched' },
+{ oid => '9998', descr => 'statistics: number of tuples fetched by idxscan',
+  proname => 'pg_stat_get_ind_tuples_fetched', provolatile => 's',
+  proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
+  prosrc => 'pg_stat_get_ind_tuples_fetched' },
 { oid => '1931', descr => 'statistics: number of tuples inserted',
-  proname => 'pg_stat_get_tuples_inserted', provolatile => 's',
+  proname => 'pg_stat_get_tab_tuples_inserted', provolatile => 's',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_tuples_inserted' },
+  prosrc => 'pg_stat_get_tab_tuples_inserted' },
 { oid => '1932', descr => 'statistics: number of tuples updated',
-  proname => 'pg_stat_get_tuples_updated', provolatile => 's',
+  proname => 'pg_stat_get_tab_tuples_updated', provolatile => 's',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_tuples_updated' },
+  prosrc => 'pg_stat_get_tab_tuples_updated' },
 { oid => '1933', descr => 'statistics: number of tuples deleted',
-  proname => 'pg_stat_get_tuples_deleted', provolatile => 's',
+  proname => 'pg_stat_get_tab_tuples_deleted', provolatile => 's',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_tuples_deleted' },
+  prosrc => 'pg_stat_get_tab_tuples_deleted' },
 { oid => '1972', descr => 'statistics: number of tuples hot updated',
-  proname => 'pg_stat_get_tuples_hot_updated', provolatile => 's',
+  proname => 'pg_stat_get_tab_tuples_hot_updated', provolatile => 's',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_tuples_hot_updated' },
+  prosrc => 'pg_stat_get_tab_tuples_hot_updated' },
 { oid => '6217',
   descr => 'statistics: number of tuples updated onto a new page',
-  proname => 'pg_stat_get_tuples_newpage_updated', provolatile => 's',
+  proname => 'pg_stat_get_tab_tuples_newpage_updated', provolatile => 's',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_tuples_newpage_updated' },
+  prosrc => 'pg_stat_get_tab_tuples_newpage_updated' },
 { oid => '2878', descr => 'statistics: number of live tuples',
-  proname => 'pg_stat_get_live_tuples', provolatile => 's', proparallel => 'r',
+  proname => 'pg_stat_get_tab_live_tuples', provolatile => 's', proparallel => 'r',
   prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_live_tuples' },
+  prosrc => 'pg_stat_get_tab_live_tuples' },
 { oid => '2879', descr => 'statistics: number of dead tuples',
-  proname => 'pg_stat_get_dead_tuples', provolatile => 's', proparallel => 'r',
+  proname => 'pg_stat_get_tab_dead_tuples', provolatile => 's', proparallel => 'r',
   prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_dead_tuples' },
+  prosrc => 'pg_stat_get_tab_dead_tuples' },
 { oid => '3177',
   descr => 'statistics: number of tuples changed since last analyze',
-  proname => 'pg_stat_get_mod_since_analyze', provolatile => 's',
+  proname => 'pg_stat_get_tab_mod_since_analyze', provolatile => 's',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_mod_since_analyze' },
+  prosrc => 'pg_stat_get_tab_mod_since_analyze' },
 { oid => '5053',
   descr => 'statistics: number of tuples inserted since last vacuum',
-  proname => 'pg_stat_get_ins_since_vacuum', provolatile => 's',
+  proname => 'pg_stat_get_tab_ins_since_vacuum', provolatile => 's',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_ins_since_vacuum' },
+  prosrc => 'pg_stat_get_tab_ins_since_vacuum' },
 { oid => '1934', descr => 'statistics: number of blocks fetched',
-  proname => 'pg_stat_get_blocks_fetched', provolatile => 's',
+  proname => 'pg_stat_get_tab_blocks_fetched', provolatile => 's',
+  proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
+  prosrc => 'pg_stat_get_tab_blocks_fetched' },
+{ oid => '8120', descr => 'statistics: number of blocks fetched',
+  proname => 'pg_stat_get_ind_blocks_fetched', provolatile => 's',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_blocks_fetched' },
+  prosrc => 'pg_stat_get_ind_blocks_fetched' },
 { oid => '1935', descr => 'statistics: number of blocks found in cache',
-  proname => 'pg_stat_get_blocks_hit', provolatile => 's', proparallel => 'r',
+  proname => 'pg_stat_get_tab_blocks_hit', provolatile => 's', proparallel => 'r',
   prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_blocks_hit' },
+  prosrc => 'pg_stat_get_tab_blocks_hit' },
+{ oid => '9982', descr => 'statistics: number of blocks found in cache',
+  proname => 'pg_stat_get_ind_blocks_hit', provolatile => 's', proparallel => 'r',
+  prorettype => 'int8', proargtypes => 'oid',
+  prosrc => 'pg_stat_get_ind_blocks_hit' },
 { oid => '2781', descr => 'statistics: last manual vacuum time for a table',
-  proname => 'pg_stat_get_last_vacuum_time', provolatile => 's',
+  proname => 'pg_stat_get_tab_last_vacuum_time', provolatile => 's',
   proparallel => 'r', prorettype => 'timestamptz', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_last_vacuum_time' },
+  prosrc => 'pg_stat_get_tab_last_vacuum_time' },
 { oid => '2782', descr => 'statistics: last auto vacuum time for a table',
-  proname => 'pg_stat_get_last_autovacuum_time', provolatile => 's',
+  proname => 'pg_stat_get_tab_last_autovacuum_time', provolatile => 's',
   proparallel => 'r', prorettype => 'timestamptz', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_last_autovacuum_time' },
+  prosrc => 'pg_stat_get_tab_last_autovacuum_time' },
 { oid => '2783', descr => 'statistics: last manual analyze time for a table',
-  proname => 'pg_stat_get_last_analyze_time', provolatile => 's',
+  proname => 'pg_stat_get_tab_last_analyze_time', provolatile => 's',
   proparallel => 'r', prorettype => 'timestamptz', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_last_analyze_time' },
+  prosrc => 'pg_stat_get_tab_last_analyze_time' },
 { oid => '2784', descr => 'statistics: last auto analyze time for a table',
-  proname => 'pg_stat_get_last_autoanalyze_time', provolatile => 's',
+  proname => 'pg_stat_get_tab_last_autoanalyze_time', provolatile => 's',
   proparallel => 'r', prorettype => 'timestamptz', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_last_autoanalyze_time' },
+  prosrc => 'pg_stat_get_tab_last_autoanalyze_time' },
 { oid => '3054', descr => 'statistics: number of manual vacuums for a table',
-  proname => 'pg_stat_get_vacuum_count', provolatile => 's', proparallel => 'r',
+  proname => 'pg_stat_get_tab_vacuum_count', provolatile => 's', proparallel => 'r',
   prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_vacuum_count' },
+  prosrc => 'pg_stat_get_tab_vacuum_count' },
 { oid => '3055', descr => 'statistics: number of auto vacuums for a table',
-  proname => 'pg_stat_get_autovacuum_count', provolatile => 's',
+  proname => 'pg_stat_get_tab_autovacuum_count', provolatile => 's',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_autovacuum_count' },
+  prosrc => 'pg_stat_get_tab_autovacuum_count' },
 { oid => '3056', descr => 'statistics: number of manual analyzes for a table',
-  proname => 'pg_stat_get_analyze_count', provolatile => 's',
+  proname => 'pg_stat_get_tab_analyze_count', provolatile => 's',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_analyze_count' },
+  prosrc => 'pg_stat_get_tab_analyze_count' },
 { oid => '3057', descr => 'statistics: number of auto analyzes for a table',
-  proname => 'pg_stat_get_autoanalyze_count', provolatile => 's',
+  proname => 'pg_stat_get_tab_autoanalyze_count', provolatile => 's',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_autoanalyze_count' },
+  prosrc => 'pg_stat_get_tab_autoanalyze_count' },
 { oid => '1936', descr => 'statistics: currently active backend IDs',
   proname => 'pg_stat_get_backend_idset', prorows => '100', proretset => 't',
   provolatile => 's', proparallel => 'r', prorettype => 'int4',
@@ -5795,55 +5819,65 @@
   prosrc => 'pg_stat_get_function_self_time' },
 
 { oid => '3037',
-  descr => 'statistics: number of scans done for table/index in current transaction',
-  proname => 'pg_stat_get_xact_numscans', provolatile => 'v',
+  descr => 'statistics: number of scans done for table in current transaction',
+  proname => 'pg_stat_get_tab_xact_numscans', provolatile => 'v',
+  proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
+  prosrc => 'pg_stat_get_tab_xact_numscans' },
+{ oid => '8892',
+  descr => 'statistics: number of scans done for index in current transaction',
+  proname => 'pg_stat_get_ind_xact_numscans', provolatile => 'v',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_xact_numscans' },
+  prosrc => 'pg_stat_get_ind_xact_numscans' },
 { oid => '3038',
   descr => 'statistics: number of tuples read by seqscan in current transaction',
-  proname => 'pg_stat_get_xact_tuples_returned', provolatile => 'v',
+  proname => 'pg_stat_get_tab_xact_tuples_returned', provolatile => 'v',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_xact_tuples_returned' },
+  prosrc => 'pg_stat_get_tab_xact_tuples_returned' },
 { oid => '3039',
   descr => 'statistics: number of tuples fetched by idxscan in current transaction',
-  proname => 'pg_stat_get_xact_tuples_fetched', provolatile => 'v',
+  proname => 'pg_stat_get_tab_xact_tuples_fetched', provolatile => 'v',
+  proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
+  prosrc => 'pg_stat_get_tab_xact_tuples_fetched' },
+{ oid => '9700',
+  descr => 'statistics: number of tuples fetched by idxscan in current transaction',
+  proname => 'pg_stat_get_ind_xact_tuples_fetched', provolatile => 'v',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_xact_tuples_fetched' },
+  prosrc => 'pg_stat_get_ind_xact_tuples_fetched' },
 { oid => '3040',
   descr => 'statistics: number of tuples inserted in current transaction',
-  proname => 'pg_stat_get_xact_tuples_inserted', provolatile => 'v',
+  proname => 'pg_stat_get_tab_xact_tuples_inserted', provolatile => 'v',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_xact_tuples_inserted' },
+  prosrc => 'pg_stat_get_tab_xact_tuples_inserted' },
 { oid => '3041',
   descr => 'statistics: number of tuples updated in current transaction',
-  proname => 'pg_stat_get_xact_tuples_updated', provolatile => 'v',
+  proname => 'pg_stat_get_tab_xact_tuples_updated', provolatile => 'v',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_xact_tuples_updated' },
+  prosrc => 'pg_stat_get_tab_xact_tuples_updated' },
 { oid => '3042',
   descr => 'statistics: number of tuples deleted in current transaction',
-  proname => 'pg_stat_get_xact_tuples_deleted', provolatile => 'v',
+  proname => 'pg_stat_get_tab_xact_tuples_deleted', provolatile => 'v',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_xact_tuples_deleted' },
+  prosrc => 'pg_stat_get_tab_xact_tuples_deleted' },
 { oid => '3043',
   descr => 'statistics: number of tuples hot updated in current transaction',
-  proname => 'pg_stat_get_xact_tuples_hot_updated', provolatile => 'v',
+  proname => 'pg_stat_get_tab_xact_tuples_hot_updated', provolatile => 'v',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_xact_tuples_hot_updated' },
+  prosrc => 'pg_stat_get_tab_xact_tuples_hot_updated' },
 { oid => '6218',
   descr => 'statistics: number of tuples updated onto a new page in current transaction',
-  proname => 'pg_stat_get_xact_tuples_newpage_updated', provolatile => 'v',
+  proname => 'pg_stat_get_tab_xact_tuples_newpage_updated', provolatile => 'v',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_xact_tuples_newpage_updated' },
+  prosrc => 'pg_stat_get_tab_xact_tuples_newpage_updated' },
 { oid => '3044',
   descr => 'statistics: number of blocks fetched in current transaction',
-  proname => 'pg_stat_get_xact_blocks_fetched', provolatile => 'v',
+  proname => 'pg_stat_get_tab_xact_blocks_fetched', provolatile => 'v',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_xact_blocks_fetched' },
+  prosrc => 'pg_stat_get_tab_xact_blocks_fetched' },
 { oid => '3045',
   descr => 'statistics: number of blocks found in cache in current transaction',
-  proname => 'pg_stat_get_xact_blocks_hit', provolatile => 'v',
+  proname => 'pg_stat_get_tab_xact_blocks_hit', provolatile => 'v',
   proparallel => 'r', prorettype => 'int8', proargtypes => 'oid',
-  prosrc => 'pg_stat_get_xact_blocks_hit' },
+  prosrc => 'pg_stat_get_tab_xact_blocks_hit' },
 { oid => '3046',
   descr => 'statistics: number of function calls in current transaction',
   proname => 'pg_stat_get_xact_function_calls', provolatile => 'v',
diff --git a/src/include/pgstat.h b/src/include/pgstat.h
index f95d8db0c4..0b701ee9aa 100644
--- a/src/include/pgstat.h
+++ b/src/include/pgstat.h
@@ -39,7 +39,8 @@ typedef enum PgStat_Kind
 
 	/* stats for variable-numbered objects */
 	PGSTAT_KIND_DATABASE,		/* database-wide statistics */
-	PGSTAT_KIND_RELATION,		/* per-table statistics */
+	PGSTAT_KIND_TABLE,			/* per-table statistics */
+	PGSTAT_KIND_INDEX,			/* per-index statistics */
 	PGSTAT_KIND_FUNCTION,		/* per-function statistics */
 	PGSTAT_KIND_REPLSLOT,		/* per-slot statistics */
 	PGSTAT_KIND_SUBSCRIPTION,	/* per-subscription statistics */
@@ -137,6 +138,28 @@ typedef struct PgStat_BackendSubEntry
 	PgStat_Counter sync_error_count;
 } PgStat_BackendSubEntry;
 
+/* ----------
+ * PgStat_IndexCounts			The actual per-index counts kept by a backend
+ *
+ * This struct should contain only actual event counters, because we memcmp
+ * it against zeroes to detect whether there are any stats updates to apply.
+ * It is a component of PgStat_IndexStatus (within-backend state).
+ *
+ * tuples_returned is the number of index entries returned by
+ * the index AM, while tuples_fetched is the number of tuples successfully
+ * fetched by heap_fetch under the control of simple indexscans for this index.
+ * ----------
+ */
+typedef struct PgStat_IndexCounts
+{
+	PgStat_Counter numscans;
+
+	PgStat_Counter tuples_returned;
+	PgStat_Counter tuples_fetched;
+	PgStat_Counter blocks_fetched;
+	PgStat_Counter blocks_hit;
+} PgStat_IndexCounts;
+
 /* ----------
  * PgStat_TableCounts			The actual per-table counts kept by a backend
  *
@@ -144,12 +167,9 @@ typedef struct PgStat_BackendSubEntry
  * it against zeroes to detect whether there are any stats updates to apply.
  * It is a component of PgStat_TableStatus (within-backend state).
  *
- * Note: for a table, tuples_returned is the number of tuples successfully
+ * Note: tuples_returned is the number of tuples successfully
  * fetched by heap_getnext, while tuples_fetched is the number of tuples
  * successfully fetched by heap_fetch under the control of bitmap indexscans.
- * For an index, tuples_returned is the number of index entries returned by
- * the index AM, while tuples_fetched is the number of tuples successfully
- * fetched by heap_fetch under the control of simple indexscans for this index.
  *
  * tuples_inserted/updated/deleted/hot_updated/newpage_updated count attempted
  * actions, regardless of whether the transaction committed.  delta_live_tuples,
@@ -203,6 +223,23 @@ typedef struct PgStat_TableStatus
 	Relation	relation;		/* rel that is using this entry */
 } PgStat_TableStatus;
 
+/* ----------
+ * PgStat_IndexStatus			Per-index status within a backend
+ *
+ * Many of the event counters are nontransactional, ie, we count events
+ * in committed and aborted transactions alike.  For these, we just count
+ * directly in the PgStat_IndexStatus.
+ * ----------
+ */
+typedef struct PgStat_IndexStatus
+{
+	Oid			r_id;			/* relation's OID */
+	bool		r_shared;		/* is it a shared catalog? */
+	struct PgStat_IndexXactStatus *trans;	/* lowest subxact's counts */
+	PgStat_IndexCounts counts;	/* event counts to be sent */
+	Relation	relation;		/* rel that is using this entry */
+} PgStat_IndexStatus;
+
 /* ----------
  * PgStat_TableXactStatus		Per-table, per-subtransaction status
  * ----------
@@ -227,6 +264,29 @@ typedef struct PgStat_TableXactStatus
 } PgStat_TableXactStatus;
 
 
+/* ----------
+ * PgStat_IndexXactStatus		Per-index, per-subtransaction status
+ * ----------
+ */
+typedef struct PgStat_IndexXactStatus
+{
+	PgStat_Counter tuples_inserted; /* tuples inserted in (sub)xact */
+	PgStat_Counter tuples_updated;	/* tuples updated in (sub)xact */
+	PgStat_Counter tuples_deleted;	/* tuples deleted in (sub)xact */
+	bool		truncdropped;	/* relation truncated/dropped in this
+								 * (sub)xact */
+	/* tuples i/u/d prior to truncate/drop */
+	PgStat_Counter inserted_pre_truncdrop;
+	PgStat_Counter updated_pre_truncdrop;
+	PgStat_Counter deleted_pre_truncdrop;
+	int			nest_level;		/* subtransaction nest level */
+	/* links to other structs for same relation: */
+	struct PgStat_IndexXactStatus *upper;	/* next higher subxact if any */
+	PgStat_IndexStatus *parent; /* per-table status */
+	/* structs of same subxact level are linked here: */
+	struct PgStat_IndexXactStatus *next;	/* next of same subxact */
+} PgStat_IndexXactStatus;
+
 /* ------------------------------------------------------------
  * Data structures on disk and in shared memory follow
  *
@@ -425,6 +485,17 @@ typedef struct PgStat_StatTabEntry
 	PgStat_Counter autoanalyze_count;
 } PgStat_StatTabEntry;
 
+typedef struct PgStat_StatIndEntry
+{
+	PgStat_Counter numscans;
+	TimestampTz lastscan;
+
+	PgStat_Counter tuples_returned;
+	PgStat_Counter tuples_fetched;
+	PgStat_Counter blocks_fetched;
+	PgStat_Counter blocks_hit;
+} PgStat_StatIndEntry;
+
 typedef struct PgStat_WalStats
 {
 	PgStat_Counter wal_records;
@@ -576,16 +647,15 @@ extern PgStat_FunctionCounts *find_funcstat_entry(Oid func_id);
 
 
 /*
- * Functions in pgstat_relation.c
+ * Functions in pgstat_table.c
  */
 
-extern void pgstat_create_relation(Relation rel);
-extern void pgstat_drop_relation(Relation rel);
-extern void pgstat_copy_relation_stats(Relation dst, Relation src);
+extern void pgstat_create_table(Relation rel);
+extern void pgstat_drop_table(Relation rel);
+extern void pgstat_init_table(Relation rel);
 
-extern void pgstat_init_relation(Relation rel);
-extern void pgstat_assoc_relation(Relation rel);
-extern void pgstat_unlink_relation(Relation rel);
+extern void pgstat_assoc_table(Relation rel);
+extern void pgstat_unlink_table(Relation rel);
 
 extern void pgstat_report_vacuum(Oid tableoid, bool shared,
 								 PgStat_Counter livetuples, PgStat_Counter deadtuples);
@@ -593,51 +663,82 @@ extern void pgstat_report_analyze(Relation rel,
 								  PgStat_Counter livetuples, PgStat_Counter deadtuples,
 								  bool resetcounter);
 
+/*
+ * Functions in pgstat_index.c
+ */
+
+extern void pgstat_create_index(Relation rel);
+extern void pgstat_drop_index(Relation rel);
+extern void pgstat_copy_index_stats(Relation dst, Relation src);
+extern void pgstat_init_index(Relation rel);
+extern void pgstat_assoc_index(Relation rel);
+extern void pgstat_unlink_index(Relation rel);
+
 /*
  * If stats are enabled, but pending data hasn't been prepared yet, call
- * pgstat_assoc_relation() to do so. See its comment for why this is done
- * separately from pgstat_init_relation().
+ * pgstat_assoc_table() / pgstat_assoc_index() to do so.
+ * See their comment for why this is done separately from pgstat_init_table()
+ * and pgstat_init_index().
  */
-#define pgstat_should_count_relation(rel)                           \
-	(likely((rel)->pgstat_info != NULL) ? true :                    \
-	 ((rel)->pgstat_enabled ? pgstat_assoc_relation(rel), true : false))
+#define pgstat_should_count_table(rel)                           \
+	(likely((rel)->pgstattab_info != NULL) ? true :              \
+	 ((rel)->pgstat_enabled ? pgstat_assoc_table(rel), true : false))
+
+#define pgstat_should_count_index(rel)                           \
+	(likely((rel)->pgstatind_info != NULL) ? true :              \
+	 ((rel)->pgstat_enabled ? pgstat_assoc_index(rel), true : false))
 
 /* nontransactional event counts are simple enough to inline */
 
 #define pgstat_count_heap_scan(rel)									\
 	do {															\
-		if (pgstat_should_count_relation(rel))						\
-			(rel)->pgstat_info->counts.numscans++;					\
+		if (pgstat_should_count_table(rel))							\
+			(rel)->pgstattab_info->counts.numscans++;				\
 	} while (0)
 #define pgstat_count_heap_getnext(rel)								\
 	do {															\
-		if (pgstat_should_count_relation(rel))						\
-			(rel)->pgstat_info->counts.tuples_returned++;			\
+		if (pgstat_should_count_table(rel))							\
+			(rel)->pgstattab_info->counts.tuples_returned++;		\
 	} while (0)
 #define pgstat_count_heap_fetch(rel)								\
 	do {															\
-		if (pgstat_should_count_relation(rel))						\
-			(rel)->pgstat_info->counts.tuples_fetched++;			\
+		if (pgstat_should_count_table(rel))							\
+			(rel)->pgstattab_info->counts.tuples_fetched++;			\
+	} while (0)
+#define pgstat_count_index_fetch(rel)								\
+	do {															\
+		if (pgstat_should_count_index(rel))							\
+			(rel)->pgstatind_info->counts.tuples_fetched++;			\
 	} while (0)
 #define pgstat_count_index_scan(rel)								\
 	do {															\
-		if (pgstat_should_count_relation(rel))						\
-			(rel)->pgstat_info->counts.numscans++;					\
+		if (pgstat_should_count_index(rel))							\
+			(rel)->pgstatind_info->counts.numscans++;				\
 	} while (0)
 #define pgstat_count_index_tuples(rel, n)							\
 	do {															\
-		if (pgstat_should_count_relation(rel))						\
-			(rel)->pgstat_info->counts.tuples_returned += (n);		\
+		if (pgstat_should_count_index(rel))							\
+			(rel)->pgstatind_info->counts.tuples_returned += (n);	\
+	} while (0)
+#define pgstat_count_table_buffer_read(rel)							\
+	do {															\
+		if (pgstat_should_count_table(rel))							\
+			(rel)->pgstattab_info->counts.blocks_fetched++;			\
+	} while (0)
+#define pgstat_count_index_buffer_read(rel)							\
+	do {															\
+		if (pgstat_should_count_index(rel))							\
+			(rel)->pgstatind_info->counts.blocks_fetched++;			\
 	} while (0)
-#define pgstat_count_buffer_read(rel)								\
+#define pgstat_count_table_buffer_hit(rel)							\
 	do {															\
-		if (pgstat_should_count_relation(rel))						\
-			(rel)->pgstat_info->counts.blocks_fetched++;			\
+		if (pgstat_should_count_table(rel))							\
+			(rel)->pgstattab_info->counts.blocks_hit++;				\
 	} while (0)
-#define pgstat_count_buffer_hit(rel)								\
+#define pgstat_count_index_buffer_hit(rel)							\
 	do {															\
-		if (pgstat_should_count_relation(rel))						\
-			(rel)->pgstat_info->counts.blocks_hit++;				\
+		if (pgstat_should_count_index(rel))							\
+			(rel)->pgstatind_info->counts.blocks_hit++;				\
 	} while (0)
 
 extern void pgstat_count_heap_insert(Relation rel, PgStat_Counter n);
@@ -655,6 +756,10 @@ extern PgStat_StatTabEntry *pgstat_fetch_stat_tabentry(Oid relid);
 extern PgStat_StatTabEntry *pgstat_fetch_stat_tabentry_ext(bool shared,
 														   Oid reloid);
 extern PgStat_TableStatus *find_tabstat_entry(Oid rel_id);
+extern PgStat_IndexStatus *find_indstat_entry(Oid rel_id);
+extern PgStat_StatIndEntry *pgstat_fetch_stat_indentry(Oid relid);
+extern PgStat_StatIndEntry *pgstat_fetch_stat_indentry_ext(bool shared,
+														   Oid relid);
 
 
 /*
diff --git a/src/include/utils/pgstat_internal.h b/src/include/utils/pgstat_internal.h
index 60fbf9394b..c9336418b4 100644
--- a/src/include/utils/pgstat_internal.h
+++ b/src/include/utils/pgstat_internal.h
@@ -370,11 +370,17 @@ typedef struct PgStatShared_Database
 	PgStat_StatDBEntry stats;
 } PgStatShared_Database;
 
-typedef struct PgStatShared_Relation
+typedef struct PgStatShared_Table
 {
 	PgStatShared_Common header;
 	PgStat_StatTabEntry stats;
-} PgStatShared_Relation;
+} PgStatShared_Table;
+
+typedef struct PgStatShared_Index
+{
+	PgStatShared_Common header;
+	PgStat_StatIndEntry stats;
+} PgStatShared_Index;
 
 typedef struct PgStatShared_Function
 {
@@ -573,16 +579,23 @@ extern void pgstat_io_snapshot_cb(void);
 
 
 /*
- * Functions in pgstat_relation.c
+ * Functions in pgstat_table.c
  */
 
-extern void AtEOXact_PgStat_Relations(PgStat_SubXactStatus *xact_state, bool isCommit);
-extern void AtEOSubXact_PgStat_Relations(PgStat_SubXactStatus *xact_state, bool isCommit, int nestDepth);
-extern void AtPrepare_PgStat_Relations(PgStat_SubXactStatus *xact_state);
-extern void PostPrepare_PgStat_Relations(PgStat_SubXactStatus *xact_state);
+extern void AtEOXact_PgStat_Tables(PgStat_SubXactStatus *xact_state, bool isCommit);
+extern void AtEOSubXact_PgStat_Tables(PgStat_SubXactStatus *xact_state, bool isCommit, int nestDepth);
+extern void AtPrepare_PgStat_Tables(PgStat_SubXactStatus *xact_state);
+extern void PostPrepare_PgStat_Tables(PgStat_SubXactStatus *xact_state);
+
+extern bool pgstat_table_flush_cb(PgStat_EntryRef *entry_ref, bool nowait);
+extern void pgstat_table_delete_pending_cb(PgStat_EntryRef *entry_ref);
+
+/*
+ * Functions in pgstat_index.c
+ */
 
-extern bool pgstat_relation_flush_cb(PgStat_EntryRef *entry_ref, bool nowait);
-extern void pgstat_relation_delete_pending_cb(PgStat_EntryRef *entry_ref);
+extern bool pgstat_index_flush_cb(PgStat_EntryRef *entry_ref, bool nowait);
+extern void pgstat_index_delete_pending_cb(PgStat_EntryRef *entry_ref);
 
 
 /*
diff --git a/src/include/utils/rel.h b/src/include/utils/rel.h
index 0ad613c4b8..fd56e737c5 100644
--- a/src/include/utils/rel.h
+++ b/src/include/utils/rel.h
@@ -251,7 +251,10 @@ typedef struct RelationData
 
 	bool		pgstat_enabled; /* should relation stats be counted */
 	/* use "struct" here to avoid needing to include pgstat.h: */
-	struct PgStat_TableStatus *pgstat_info; /* statistics collection area */
+	/* table's statistics collection area */
+	struct PgStat_TableStatus *pgstattab_info;
+	/* Index's statistics collection area */
+	struct PgStat_IndexStatus *pgstatind_info;
 } RelationData;
 
 
diff --git a/src/test/isolation/expected/stats.out b/src/test/isolation/expected/stats.out
index 61b5a710ec..494adee88b 100644
--- a/src/test/isolation/expected/stats.out
+++ b/src/test/isolation/expected/stats.out
@@ -2166,14 +2166,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2210,14 +2210,14 @@ step s2_table_update_k1: UPDATE test_stat_tab SET value = value + 1 WHERE key =
 step s1_table_drop: DROP TABLE test_stat_tab;
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2236,14 +2236,14 @@ pg_stat_force_next_flush
 step s1_track_counts_off: SET track_counts = off;
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2275,14 +2275,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2314,14 +2314,14 @@ pg_stat_force_next_flush
 step s1_track_counts_off: SET track_counts = off;
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2369,14 +2369,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2401,14 +2401,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2463,14 +2463,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2495,14 +2495,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2558,14 +2558,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2627,14 +2627,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2688,14 +2688,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2755,14 +2755,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2795,14 +2795,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2841,14 +2841,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2881,14 +2881,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2927,14 +2927,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2968,14 +2968,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -3015,14 +3015,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
diff --git a/src/test/isolation/expected/stats_1.out b/src/test/isolation/expected/stats_1.out
index 3854320106..7981747732 100644
--- a/src/test/isolation/expected/stats_1.out
+++ b/src/test/isolation/expected/stats_1.out
@@ -2174,14 +2174,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2218,14 +2218,14 @@ step s2_table_update_k1: UPDATE test_stat_tab SET value = value + 1 WHERE key =
 step s1_table_drop: DROP TABLE test_stat_tab;
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2244,14 +2244,14 @@ pg_stat_force_next_flush
 step s1_track_counts_off: SET track_counts = off;
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2283,14 +2283,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2322,14 +2322,14 @@ pg_stat_force_next_flush
 step s1_track_counts_off: SET track_counts = off;
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2377,14 +2377,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2409,14 +2409,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2471,14 +2471,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2503,14 +2503,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2566,14 +2566,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2635,14 +2635,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2698,14 +2698,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2767,14 +2767,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2809,14 +2809,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2857,14 +2857,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2899,14 +2899,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2947,14 +2947,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -2990,14 +2990,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
@@ -3039,14 +3039,14 @@ pg_stat_force_next_flush
 
 step s1_table_stats: 
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
         pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 
diff --git a/src/test/isolation/specs/stats.spec b/src/test/isolation/specs/stats.spec
index 5b922d788c..53d78f7db1 100644
--- a/src/test/isolation/specs/stats.spec
+++ b/src/test/isolation/specs/stats.spec
@@ -93,14 +93,14 @@ step s1_table_drop { DROP TABLE test_stat_tab; }
 
 step s1_table_stats {
     SELECT
-        pg_stat_get_numscans(tso.oid) AS seq_scan,
-        pg_stat_get_tuples_returned(tso.oid) AS seq_tup_read,
-        pg_stat_get_tuples_inserted(tso.oid) AS n_tup_ins,
-        pg_stat_get_tuples_updated(tso.oid) AS n_tup_upd,
-        pg_stat_get_tuples_deleted(tso.oid) AS n_tup_del,
-        pg_stat_get_live_tuples(tso.oid) AS n_live_tup,
-        pg_stat_get_dead_tuples(tso.oid) AS n_dead_tup,
-        pg_stat_get_vacuum_count(tso.oid) AS vacuum_count
+        pg_stat_get_tab_numscans(tso.oid) AS seq_scan,
+        pg_stat_get_tab_tuples_returned(tso.oid) AS seq_tup_read,
+        pg_stat_get_tab_tuples_inserted(tso.oid) AS n_tup_ins,
+        pg_stat_get_tab_tuples_updated(tso.oid) AS n_tup_upd,
+        pg_stat_get_tab_tuples_deleted(tso.oid) AS n_tup_del,
+        pg_stat_get_tab_live_tuples(tso.oid) AS n_live_tup,
+        pg_stat_get_tab_dead_tuples(tso.oid) AS n_dead_tup,
+        pg_stat_get_tab_vacuum_count(tso.oid) AS vacuum_count
     FROM test_stat_oid AS tso
     WHERE tso.name = 'test_stat_tab'
 }
diff --git a/src/test/recovery/t/029_stats_restart.pl b/src/test/recovery/t/029_stats_restart.pl
index f6368ab1d3..ecfd359b74 100644
--- a/src/test/recovery/t/029_stats_restart.pl
+++ b/src/test/recovery/t/029_stats_restart.pl
@@ -43,8 +43,8 @@ my $sect = "initial";
 is(have_stats('database', $dboid, 0), 't', "$sect: db stats do exist");
 is(have_stats('function', $dboid, $funcoid),
 	't', "$sect: function stats do exist");
-is(have_stats('relation', $dboid, $tableoid),
-	't', "$sect: relation stats do exist");
+is(have_stats('table', $dboid, $tableoid),
+	't', "$sect: table stats do exist");
 
 # regular shutdown
 $node->stop();
@@ -67,8 +67,8 @@ $sect = "copy";
 is(have_stats('database', $dboid, 0), 't', "$sect: db stats do exist");
 is(have_stats('function', $dboid, $funcoid),
 	't', "$sect: function stats do exist");
-is(have_stats('relation', $dboid, $tableoid),
-	't', "$sect: relation stats do exist");
+is(have_stats('table', $dboid, $tableoid),
+	't', "$sect: table stats do exist");
 
 $node->stop('immediate');
 
@@ -84,8 +84,8 @@ $sect = "post immediate";
 is(have_stats('database', $dboid, 0), 'f', "$sect: db stats do not exist");
 is(have_stats('function', $dboid, $funcoid),
 	'f', "$sect: function stats do exist");
-is(have_stats('relation', $dboid, $tableoid),
-	'f', "$sect: relation stats do not exist");
+is(have_stats('table', $dboid, $tableoid),
+	'f', "$sect: table stats do not exist");
 
 # get rid of backup statsfile
 unlink $statsfile or die "cannot unlink $statsfile $!";
@@ -98,8 +98,8 @@ $sect = "post immediate, new";
 is(have_stats('database', $dboid, 0), 't', "$sect: db stats do exist");
 is(have_stats('function', $dboid, $funcoid),
 	't', "$sect: function stats do exist");
-is(have_stats('relation', $dboid, $tableoid),
-	't', "$sect: relation stats do exist");
+is(have_stats('table', $dboid, $tableoid),
+	't', "$sect: table stats do exist");
 
 # regular shutdown
 $node->stop();
@@ -117,8 +117,7 @@ $sect = "invalid_overwrite";
 is(have_stats('database', $dboid, 0), 'f', "$sect: db stats do not exist");
 is(have_stats('function', $dboid, $funcoid),
 	'f', "$sect: function stats do not exist");
-is(have_stats('relation', $dboid, $tableoid),
-	'f', "$sect: relation stats do not exist");
+is(have_stats('table', $dboid, $tableoid), 'f', "$sect: table do not exist");
 
 
 ## check invalid stats file starting with valid contents, but followed by
@@ -133,8 +132,8 @@ $sect = "invalid_append";
 is(have_stats('database', $dboid, 0), 'f', "$sect: db stats do not exist");
 is(have_stats('function', $dboid, $funcoid),
 	'f', "$sect: function stats do not exist");
-is(have_stats('relation', $dboid, $tableoid),
-	'f', "$sect: relation stats do not exist");
+is(have_stats('table', $dboid, $tableoid),
+	'f', "$sect: table stats do not exist");
 
 
 ## checks related to stats persistency around restarts and resets
diff --git a/src/test/recovery/t/030_stats_cleanup_replica.pl b/src/test/recovery/t/030_stats_cleanup_replica.pl
index 51495aebcd..042ae55f17 100644
--- a/src/test/recovery/t/030_stats_cleanup_replica.pl
+++ b/src/test/recovery/t/030_stats_cleanup_replica.pl
@@ -179,7 +179,7 @@ sub test_standby_func_tab_stats_status
 	my %stats;
 
 	$stats{rel} = $node_standby->safe_psql($connect_db,
-		"SELECT pg_stat_have_stats('relation', $dboid, $tableoid)");
+		"SELECT pg_stat_have_stats('table', $dboid, $tableoid)");
 	$stats{func} = $node_standby->safe_psql($connect_db,
 		"SELECT pg_stat_have_stats('function', $dboid, $funcoid)");
 
diff --git a/src/test/regress/expected/rules.out b/src/test/regress/expected/rules.out
index 1442c43d9c..a4cace6ef3 100644
--- a/src/test/regress/expected/rules.out
+++ b/src/test/regress/expected/rules.out
@@ -1768,10 +1768,10 @@ pg_stat_all_indexes| SELECT c.oid AS relid,
     n.nspname AS schemaname,
     c.relname,
     i.relname AS indexrelname,
-    pg_stat_get_numscans(i.oid) AS idx_scan,
-    pg_stat_get_lastscan(i.oid) AS last_idx_scan,
-    pg_stat_get_tuples_returned(i.oid) AS idx_tup_read,
-    pg_stat_get_tuples_fetched(i.oid) AS idx_tup_fetch
+    pg_stat_get_ind_numscans(i.oid) AS idx_scan,
+    pg_stat_get_ind_lastscan(i.oid) AS last_idx_scan,
+    pg_stat_get_ind_tuples_returned(i.oid) AS idx_tup_read,
+    pg_stat_get_ind_tuples_fetched(i.oid) AS idx_tup_fetch
    FROM (((pg_class c
      JOIN pg_index x ON ((c.oid = x.indrelid)))
      JOIN pg_class i ON ((i.oid = x.indexrelid)))
@@ -1780,29 +1780,29 @@ pg_stat_all_indexes| SELECT c.oid AS relid,
 pg_stat_all_tables| SELECT c.oid AS relid,
     n.nspname AS schemaname,
     c.relname,
-    pg_stat_get_numscans(c.oid) AS seq_scan,
-    pg_stat_get_lastscan(c.oid) AS last_seq_scan,
-    pg_stat_get_tuples_returned(c.oid) AS seq_tup_read,
-    (sum(pg_stat_get_numscans(i.indexrelid)))::bigint AS idx_scan,
-    max(pg_stat_get_lastscan(i.indexrelid)) AS last_idx_scan,
-    ((sum(pg_stat_get_tuples_fetched(i.indexrelid)))::bigint + pg_stat_get_tuples_fetched(c.oid)) AS idx_tup_fetch,
-    pg_stat_get_tuples_inserted(c.oid) AS n_tup_ins,
-    pg_stat_get_tuples_updated(c.oid) AS n_tup_upd,
-    pg_stat_get_tuples_deleted(c.oid) AS n_tup_del,
-    pg_stat_get_tuples_hot_updated(c.oid) AS n_tup_hot_upd,
-    pg_stat_get_tuples_newpage_updated(c.oid) AS n_tup_newpage_upd,
-    pg_stat_get_live_tuples(c.oid) AS n_live_tup,
-    pg_stat_get_dead_tuples(c.oid) AS n_dead_tup,
-    pg_stat_get_mod_since_analyze(c.oid) AS n_mod_since_analyze,
-    pg_stat_get_ins_since_vacuum(c.oid) AS n_ins_since_vacuum,
-    pg_stat_get_last_vacuum_time(c.oid) AS last_vacuum,
-    pg_stat_get_last_autovacuum_time(c.oid) AS last_autovacuum,
-    pg_stat_get_last_analyze_time(c.oid) AS last_analyze,
-    pg_stat_get_last_autoanalyze_time(c.oid) AS last_autoanalyze,
-    pg_stat_get_vacuum_count(c.oid) AS vacuum_count,
-    pg_stat_get_autovacuum_count(c.oid) AS autovacuum_count,
-    pg_stat_get_analyze_count(c.oid) AS analyze_count,
-    pg_stat_get_autoanalyze_count(c.oid) AS autoanalyze_count
+    pg_stat_get_tab_numscans(c.oid) AS seq_scan,
+    pg_stat_get_tab_lastscan(c.oid) AS last_seq_scan,
+    pg_stat_get_tab_tuples_returned(c.oid) AS seq_tup_read,
+    (sum(pg_stat_get_ind_numscans(i.indexrelid)))::bigint AS idx_scan,
+    max(pg_stat_get_ind_lastscan(i.indexrelid)) AS last_idx_scan,
+    ((sum(pg_stat_get_ind_tuples_fetched(i.indexrelid)))::bigint + pg_stat_get_tab_tuples_fetched(c.oid)) AS idx_tup_fetch,
+    pg_stat_get_tab_tuples_inserted(c.oid) AS n_tup_ins,
+    pg_stat_get_tab_tuples_updated(c.oid) AS n_tup_upd,
+    pg_stat_get_tab_tuples_deleted(c.oid) AS n_tup_del,
+    pg_stat_get_tab_tuples_hot_updated(c.oid) AS n_tup_hot_upd,
+    pg_stat_get_tab_tuples_newpage_updated(c.oid) AS n_tup_newpage_upd,
+    pg_stat_get_tab_live_tuples(c.oid) AS n_live_tup,
+    pg_stat_get_tab_dead_tuples(c.oid) AS n_dead_tup,
+    pg_stat_get_tab_mod_since_analyze(c.oid) AS n_mod_since_analyze,
+    pg_stat_get_tab_ins_since_vacuum(c.oid) AS n_ins_since_vacuum,
+    pg_stat_get_tab_last_vacuum_time(c.oid) AS last_vacuum,
+    pg_stat_get_tab_last_autovacuum_time(c.oid) AS last_autovacuum,
+    pg_stat_get_tab_last_analyze_time(c.oid) AS last_analyze,
+    pg_stat_get_tab_last_autoanalyze_time(c.oid) AS last_autoanalyze,
+    pg_stat_get_tab_vacuum_count(c.oid) AS vacuum_count,
+    pg_stat_get_tab_autovacuum_count(c.oid) AS autovacuum_count,
+    pg_stat_get_tab_analyze_count(c.oid) AS analyze_count,
+    pg_stat_get_tab_autoanalyze_count(c.oid) AS autoanalyze_count
    FROM ((pg_class c
      LEFT JOIN pg_index i ON ((c.oid = i.indrelid)))
      LEFT JOIN pg_namespace n ON ((n.oid = c.relnamespace)))
@@ -2252,15 +2252,15 @@ pg_stat_wal_receiver| SELECT pid,
 pg_stat_xact_all_tables| SELECT c.oid AS relid,
     n.nspname AS schemaname,
     c.relname,
-    pg_stat_get_xact_numscans(c.oid) AS seq_scan,
-    pg_stat_get_xact_tuples_returned(c.oid) AS seq_tup_read,
-    (sum(pg_stat_get_xact_numscans(i.indexrelid)))::bigint AS idx_scan,
-    ((sum(pg_stat_get_xact_tuples_fetched(i.indexrelid)))::bigint + pg_stat_get_xact_tuples_fetched(c.oid)) AS idx_tup_fetch,
-    pg_stat_get_xact_tuples_inserted(c.oid) AS n_tup_ins,
-    pg_stat_get_xact_tuples_updated(c.oid) AS n_tup_upd,
-    pg_stat_get_xact_tuples_deleted(c.oid) AS n_tup_del,
-    pg_stat_get_xact_tuples_hot_updated(c.oid) AS n_tup_hot_upd,
-    pg_stat_get_xact_tuples_newpage_updated(c.oid) AS n_tup_newpage_upd
+    pg_stat_get_tab_xact_numscans(c.oid) AS seq_scan,
+    pg_stat_get_tab_xact_tuples_returned(c.oid) AS seq_tup_read,
+    (sum(pg_stat_get_ind_xact_numscans(i.indexrelid)))::bigint AS idx_scan,
+    ((sum(pg_stat_get_ind_xact_tuples_fetched(i.indexrelid)))::bigint + pg_stat_get_tab_xact_tuples_fetched(c.oid)) AS idx_tup_fetch,
+    pg_stat_get_tab_xact_tuples_inserted(c.oid) AS n_tup_ins,
+    pg_stat_get_tab_xact_tuples_updated(c.oid) AS n_tup_upd,
+    pg_stat_get_tab_xact_tuples_deleted(c.oid) AS n_tup_del,
+    pg_stat_get_tab_xact_tuples_hot_updated(c.oid) AS n_tup_hot_upd,
+    pg_stat_get_tab_xact_tuples_newpage_updated(c.oid) AS n_tup_newpage_upd
    FROM ((pg_class c
      LEFT JOIN pg_index i ON ((c.oid = i.indrelid)))
      LEFT JOIN pg_namespace n ON ((n.oid = c.relnamespace)))
@@ -2308,8 +2308,8 @@ pg_statio_all_indexes| SELECT c.oid AS relid,
     n.nspname AS schemaname,
     c.relname,
     i.relname AS indexrelname,
-    (pg_stat_get_blocks_fetched(i.oid) - pg_stat_get_blocks_hit(i.oid)) AS idx_blks_read,
-    pg_stat_get_blocks_hit(i.oid) AS idx_blks_hit
+    (pg_stat_get_ind_blocks_fetched(i.oid) - pg_stat_get_ind_blocks_hit(i.oid)) AS idx_blks_read,
+    pg_stat_get_ind_blocks_hit(i.oid) AS idx_blks_hit
    FROM (((pg_class c
      JOIN pg_index x ON ((c.oid = x.indrelid)))
      JOIN pg_class i ON ((i.oid = x.indexrelid)))
@@ -2318,31 +2318,31 @@ pg_statio_all_indexes| SELECT c.oid AS relid,
 pg_statio_all_sequences| SELECT c.oid AS relid,
     n.nspname AS schemaname,
     c.relname,
-    (pg_stat_get_blocks_fetched(c.oid) - pg_stat_get_blocks_hit(c.oid)) AS blks_read,
-    pg_stat_get_blocks_hit(c.oid) AS blks_hit
+    (pg_stat_get_ind_blocks_fetched(c.oid) - pg_stat_get_tab_blocks_hit(c.oid)) AS blks_read,
+    pg_stat_get_tab_blocks_hit(c.oid) AS blks_hit
    FROM (pg_class c
      LEFT JOIN pg_namespace n ON ((n.oid = c.relnamespace)))
   WHERE (c.relkind = 'S'::"char");
 pg_statio_all_tables| SELECT c.oid AS relid,
     n.nspname AS schemaname,
     c.relname,
-    (pg_stat_get_blocks_fetched(c.oid) - pg_stat_get_blocks_hit(c.oid)) AS heap_blks_read,
-    pg_stat_get_blocks_hit(c.oid) AS heap_blks_hit,
+    (pg_stat_get_tab_blocks_fetched(c.oid) - pg_stat_get_tab_blocks_hit(c.oid)) AS heap_blks_read,
+    pg_stat_get_tab_blocks_hit(c.oid) AS heap_blks_hit,
     i.idx_blks_read,
     i.idx_blks_hit,
-    (pg_stat_get_blocks_fetched(t.oid) - pg_stat_get_blocks_hit(t.oid)) AS toast_blks_read,
-    pg_stat_get_blocks_hit(t.oid) AS toast_blks_hit,
+    (pg_stat_get_tab_blocks_fetched(t.oid) - pg_stat_get_tab_blocks_hit(t.oid)) AS toast_blks_read,
+    pg_stat_get_tab_blocks_hit(t.oid) AS toast_blks_hit,
     x.idx_blks_read AS tidx_blks_read,
     x.idx_blks_hit AS tidx_blks_hit
    FROM ((((pg_class c
      LEFT JOIN pg_class t ON ((c.reltoastrelid = t.oid)))
      LEFT JOIN pg_namespace n ON ((n.oid = c.relnamespace)))
-     LEFT JOIN LATERAL ( SELECT (sum((pg_stat_get_blocks_fetched(pg_index.indexrelid) - pg_stat_get_blocks_hit(pg_index.indexrelid))))::bigint AS idx_blks_read,
-            (sum(pg_stat_get_blocks_hit(pg_index.indexrelid)))::bigint AS idx_blks_hit
+     LEFT JOIN LATERAL ( SELECT (sum((pg_stat_get_ind_blocks_fetched(pg_index.indexrelid) - pg_stat_get_ind_blocks_hit(pg_index.indexrelid))))::bigint AS idx_blks_read,
+            (sum(pg_stat_get_ind_blocks_hit(pg_index.indexrelid)))::bigint AS idx_blks_hit
            FROM pg_index
           WHERE (pg_index.indrelid = c.oid)) i ON (true))
-     LEFT JOIN LATERAL ( SELECT (sum((pg_stat_get_blocks_fetched(pg_index.indexrelid) - pg_stat_get_blocks_hit(pg_index.indexrelid))))::bigint AS idx_blks_read,
-            (sum(pg_stat_get_blocks_hit(pg_index.indexrelid)))::bigint AS idx_blks_hit
+     LEFT JOIN LATERAL ( SELECT (sum((pg_stat_get_ind_blocks_fetched(pg_index.indexrelid) - pg_stat_get_ind_blocks_hit(pg_index.indexrelid))))::bigint AS idx_blks_read,
+            (sum(pg_stat_get_ind_blocks_hit(pg_index.indexrelid)))::bigint AS idx_blks_hit
            FROM pg_index
           WHERE (pg_index.indrelid = t.oid)) x ON (true))
   WHERE (c.relkind = ANY (ARRAY['r'::"char", 't'::"char", 'm'::"char"]));
diff --git a/src/test/regress/expected/stats.out b/src/test/regress/expected/stats.out
index 0694de736a..20ff012fd2 100644
--- a/src/test/regress/expected/stats.out
+++ b/src/test/regress/expected/stats.out
@@ -151,6 +151,117 @@ FROM prevstats AS pr;
 
 COMMIT;
 ----
+-- Basic tests for toast
+---
+CREATE TABLE stats_toast(stuff text);
+ALTER TABLE stats_toast ALTER COLUMN stuff SET STORAGE EXTERNAL;
+INSERT INTO stats_toast VALUES (repeat('a',1000000));
+SELECT count(*) FROM stats_toast WHERE stuff like '%a%';
+ count 
+-------
+     1
+(1 row)
+
+-- ensure pending stats are flushed
+SELECT pg_stat_force_next_flush();
+ pg_stat_force_next_flush 
+--------------------------
+ 
+(1 row)
+
+select toast_blks_read >= 0 as toast_blks_read, toast_blks_hit > 0 as toast_blks_hit,
+       tidx_blks_read >= 0 as tidx_blks_read, tidx_blks_hit > 0 as tidx_blks_hit,
+       toast_blks_hit >= toast_blks_read,
+       tidx_blks_hit >= tidx_blks_read
+   from pg_statio_all_tables
+ where relname = 'stats_toast';
+ toast_blks_read | toast_blks_hit | tidx_blks_read | tidx_blks_hit | ?column? | ?column? 
+-----------------+----------------+----------------+---------------+----------+----------
+ t               | t              | t              | t             | t        | t
+(1 row)
+
+----
+-- Basic tests for partition
+---
+CREATE TABLE stats_partition (
+    id1         int not null,
+    id2         int not null
+) PARTITION BY RANGE (id2);
+CREATE TABLE stats_partition_5 PARTITION OF stats_partition
+    FOR VALUES FROM (0) TO (5);
+CREATE TABLE stats_partition_20000 PARTITION OF stats_partition
+    FOR VALUES FROM (5) TO (20001);
+CREATE INDEX ON stats_partition (id2);
+insert into stats_partition select a,a from generate_series(0,20000) a;
+SET enable_seqscan TO off;
+SET enable_bitmapscan TO off;
+select * from stats_partition where id2 = 2000;
+ id1  | id2  
+------+------
+ 2000 | 2000
+(1 row)
+
+select * from stats_partition where id2 = 2;
+ id1 | id2 
+-----+-----
+   2 |   2
+(1 row)
+
+-- ensure pending stats are flushed
+SELECT pg_stat_force_next_flush();
+ pg_stat_force_next_flush 
+--------------------------
+ 
+(1 row)
+
+select relname, indexrelname, idx_scan > 0 as idx_scan, idx_tup_read > 0 as idx_tup_read,
+       idx_tup_fetch > 0 as idx_tup_fetch
+   from pg_stat_all_indexes
+ where relname like '%stats_partition%'
+ order by relname COLLATE "C";
+        relname        |         indexrelname          | idx_scan | idx_tup_read | idx_tup_fetch 
+-----------------------+-------------------------------+----------+--------------+---------------
+ stats_partition_20000 | stats_partition_20000_id2_idx | t        | t            | t
+ stats_partition_5     | stats_partition_5_id2_idx     | t        | t            | t
+(2 rows)
+
+select relname, indexrelname, idx_blks_read >= 0 as idx_blks_read,
+       idx_blks_hit > 0 as idx_blks_hit
+   from pg_statio_all_indexes
+ where relname like '%stats_partition%'
+ order by relname COLLATE "C";
+        relname        |         indexrelname          | idx_blks_read | idx_blks_hit 
+-----------------------+-------------------------------+---------------+--------------
+ stats_partition_20000 | stats_partition_20000_id2_idx | t             | t
+ stats_partition_5     | stats_partition_5_id2_idx     | t             | t
+(2 rows)
+
+select relname,seq_scan > 0 as seq_scan, seq_tup_read > 0 as seq_tup_read,
+       idx_scan > 0 as idx_scan, idx_tup_fetch > 0 as idx_tup_fetch
+   from pg_stat_all_tables
+  where relname like '%stats_partition%'
+  order by relname COLLATE "C";
+        relname        | seq_scan | seq_tup_read | idx_scan | idx_tup_fetch 
+-----------------------+----------+--------------+----------+---------------
+ stats_partition       | f        | f            | f        | f
+ stats_partition_20000 | t        | f            | t        | t
+ stats_partition_5     | t        | f            | t        | t
+(3 rows)
+
+select relname, heap_blks_read >= 0 as heap_blks_read, heap_blks_hit > 0 as heap_blks_hit,
+       idx_blks_read >= 0 as idx_blks_read, idx_blks_hit > 0 as idx_blks_hit
+   from pg_statio_all_tables
+ where relname like '%stats_partition%'
+ order by relname COLLATE "C";
+        relname        | heap_blks_read | heap_blks_hit | idx_blks_read | idx_blks_hit 
+-----------------------+----------------+---------------+---------------+--------------
+ stats_partition_20000 | t              | t             | t             | t
+ stats_partition_5     | t              | t             | t             | t
+(2 rows)
+
+SET enable_seqscan TO on;
+SET enable_bitmapscan TO on;
+----
 -- Basic tests for track_functions
 ---
 CREATE FUNCTION stats_test_func1() RETURNS VOID LANGUAGE plpgsql AS $$BEGIN END;$$;
@@ -357,58 +468,58 @@ SELECT pg_stat_force_next_flush();
  
 (1 row)
 
-SELECT pg_stat_get_live_tuples(:drop_stats_test_oid);
- pg_stat_get_live_tuples 
--------------------------
-                       1
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_oid);
+ pg_stat_get_tab_live_tuples 
+-----------------------------
+                           1
 (1 row)
 
 DROP TABLE drop_stats_test;
-SELECT pg_stat_get_live_tuples(:drop_stats_test_oid);
- pg_stat_get_live_tuples 
--------------------------
-                       0
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_oid);
+ pg_stat_get_tab_live_tuples 
+-----------------------------
+                           0
 (1 row)
 
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_oid);
- pg_stat_get_xact_tuples_inserted 
-----------------------------------
-                                0
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_oid);
+ pg_stat_get_tab_xact_tuples_inserted 
+--------------------------------------
+                                    0
 (1 row)
 
 -- check that rollback protects against having stats dropped and that local
 -- modifications don't pose a problem
-SELECT pg_stat_get_live_tuples(:drop_stats_test_xact_oid);
- pg_stat_get_live_tuples 
--------------------------
-                       1
-(1 row)
-
-SELECT pg_stat_get_tuples_inserted(:drop_stats_test_xact_oid);
- pg_stat_get_tuples_inserted 
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_xact_oid);
+ pg_stat_get_tab_live_tuples 
 -----------------------------
                            1
 (1 row)
 
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_xact_oid);
- pg_stat_get_xact_tuples_inserted 
-----------------------------------
-                                0
+SELECT pg_stat_get_tab_tuples_inserted(:drop_stats_test_xact_oid);
+ pg_stat_get_tab_tuples_inserted 
+---------------------------------
+                               1
+(1 row)
+
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_xact_oid);
+ pg_stat_get_tab_xact_tuples_inserted 
+--------------------------------------
+                                    0
 (1 row)
 
 BEGIN;
 INSERT INTO drop_stats_test_xact DEFAULT VALUES;
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_xact_oid);
- pg_stat_get_xact_tuples_inserted 
-----------------------------------
-                                1
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_xact_oid);
+ pg_stat_get_tab_xact_tuples_inserted 
+--------------------------------------
+                                    1
 (1 row)
 
 DROP TABLE drop_stats_test_xact;
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_xact_oid);
- pg_stat_get_xact_tuples_inserted 
-----------------------------------
-                                0
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_xact_oid);
+ pg_stat_get_tab_xact_tuples_inserted 
+--------------------------------------
+                                    0
 (1 row)
 
 ROLLBACK;
@@ -418,44 +529,44 @@ SELECT pg_stat_force_next_flush();
  
 (1 row)
 
-SELECT pg_stat_get_live_tuples(:drop_stats_test_xact_oid);
- pg_stat_get_live_tuples 
--------------------------
-                       1
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_xact_oid);
+ pg_stat_get_tab_live_tuples 
+-----------------------------
+                           1
 (1 row)
 
-SELECT pg_stat_get_tuples_inserted(:drop_stats_test_xact_oid);
- pg_stat_get_tuples_inserted 
------------------------------
-                           2
+SELECT pg_stat_get_tab_tuples_inserted(:drop_stats_test_xact_oid);
+ pg_stat_get_tab_tuples_inserted 
+---------------------------------
+                               2
 (1 row)
 
 -- transactional drop
-SELECT pg_stat_get_live_tuples(:drop_stats_test_xact_oid);
- pg_stat_get_live_tuples 
--------------------------
-                       1
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_xact_oid);
+ pg_stat_get_tab_live_tuples 
+-----------------------------
+                           1
 (1 row)
 
-SELECT pg_stat_get_tuples_inserted(:drop_stats_test_xact_oid);
- pg_stat_get_tuples_inserted 
------------------------------
-                           2
+SELECT pg_stat_get_tab_tuples_inserted(:drop_stats_test_xact_oid);
+ pg_stat_get_tab_tuples_inserted 
+---------------------------------
+                               2
 (1 row)
 
 BEGIN;
 INSERT INTO drop_stats_test_xact DEFAULT VALUES;
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_xact_oid);
- pg_stat_get_xact_tuples_inserted 
-----------------------------------
-                                1
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_xact_oid);
+ pg_stat_get_tab_xact_tuples_inserted 
+--------------------------------------
+                                    1
 (1 row)
 
 DROP TABLE drop_stats_test_xact;
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_xact_oid);
- pg_stat_get_xact_tuples_inserted 
-----------------------------------
-                                0
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_xact_oid);
+ pg_stat_get_tab_xact_tuples_inserted 
+--------------------------------------
+                                    0
 (1 row)
 
 COMMIT;
@@ -465,42 +576,42 @@ SELECT pg_stat_force_next_flush();
  
 (1 row)
 
-SELECT pg_stat_get_live_tuples(:drop_stats_test_xact_oid);
- pg_stat_get_live_tuples 
--------------------------
-                       0
-(1 row)
-
-SELECT pg_stat_get_tuples_inserted(:drop_stats_test_xact_oid);
- pg_stat_get_tuples_inserted 
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_xact_oid);
+ pg_stat_get_tab_live_tuples 
 -----------------------------
                            0
 (1 row)
 
+SELECT pg_stat_get_tab_tuples_inserted(:drop_stats_test_xact_oid);
+ pg_stat_get_tab_tuples_inserted 
+---------------------------------
+                               0
+(1 row)
+
 -- savepoint rollback (2 levels)
-SELECT pg_stat_get_live_tuples(:drop_stats_test_subxact_oid);
- pg_stat_get_live_tuples 
--------------------------
-                       1
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_subxact_oid);
+ pg_stat_get_tab_live_tuples 
+-----------------------------
+                           1
 (1 row)
 
 BEGIN;
 INSERT INTO drop_stats_test_subxact DEFAULT VALUES;
 SAVEPOINT sp1;
 INSERT INTO drop_stats_test_subxact DEFAULT VALUES;
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_subxact_oid);
- pg_stat_get_xact_tuples_inserted 
-----------------------------------
-                                2
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_subxact_oid);
+ pg_stat_get_tab_xact_tuples_inserted 
+--------------------------------------
+                                    2
 (1 row)
 
 SAVEPOINT sp2;
 DROP TABLE drop_stats_test_subxact;
 ROLLBACK TO SAVEPOINT sp2;
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_subxact_oid);
- pg_stat_get_xact_tuples_inserted 
-----------------------------------
-                                2
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_subxact_oid);
+ pg_stat_get_tab_xact_tuples_inserted 
+--------------------------------------
+                                    2
 (1 row)
 
 COMMIT;
@@ -510,17 +621,17 @@ SELECT pg_stat_force_next_flush();
  
 (1 row)
 
-SELECT pg_stat_get_live_tuples(:drop_stats_test_subxact_oid);
- pg_stat_get_live_tuples 
--------------------------
-                       3
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_subxact_oid);
+ pg_stat_get_tab_live_tuples 
+-----------------------------
+                           3
 (1 row)
 
 -- savepoint rolback (1 level)
-SELECT pg_stat_get_live_tuples(:drop_stats_test_subxact_oid);
- pg_stat_get_live_tuples 
--------------------------
-                       3
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_subxact_oid);
+ pg_stat_get_tab_live_tuples 
+-----------------------------
+                           3
 (1 row)
 
 BEGIN;
@@ -529,17 +640,17 @@ DROP TABLE drop_stats_test_subxact;
 SAVEPOINT sp2;
 ROLLBACK TO SAVEPOINT sp1;
 COMMIT;
-SELECT pg_stat_get_live_tuples(:drop_stats_test_subxact_oid);
- pg_stat_get_live_tuples 
--------------------------
-                       3
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_subxact_oid);
+ pg_stat_get_tab_live_tuples 
+-----------------------------
+                           3
 (1 row)
 
 -- and now actually drop
-SELECT pg_stat_get_live_tuples(:drop_stats_test_subxact_oid);
- pg_stat_get_live_tuples 
--------------------------
-                       3
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_subxact_oid);
+ pg_stat_get_tab_live_tuples 
+-----------------------------
+                           3
 (1 row)
 
 BEGIN;
@@ -548,10 +659,10 @@ DROP TABLE drop_stats_test_subxact;
 SAVEPOINT sp2;
 RELEASE SAVEPOINT sp1;
 COMMIT;
-SELECT pg_stat_get_live_tuples(:drop_stats_test_subxact_oid);
- pg_stat_get_live_tuples 
--------------------------
-                       0
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_subxact_oid);
+ pg_stat_get_tab_live_tuples 
+-----------------------------
+                           0
 (1 row)
 
 DROP TABLE trunc_stats_test, trunc_stats_test1, trunc_stats_test2, trunc_stats_test3, trunc_stats_test4;
@@ -1173,21 +1284,21 @@ select a from stats_test_tab1 where a = 3;
  3
 (1 row)
 
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
  pg_stat_have_stats 
 --------------------
  t
 (1 row)
 
 -- pg_stat_have_stats returns false for dropped index with stats
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
  pg_stat_have_stats 
 --------------------
  t
 (1 row)
 
 DROP index stats_test_idx1;
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
  pg_stat_have_stats 
 --------------------
  f
@@ -1203,14 +1314,14 @@ select a from stats_test_tab1 where a = 3;
  3
 (1 row)
 
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
  pg_stat_have_stats 
 --------------------
  t
 (1 row)
 
 ROLLBACK;
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
  pg_stat_have_stats 
 --------------------
  f
@@ -1225,7 +1336,7 @@ select a from stats_test_tab1 where a = 3;
  3
 (1 row)
 
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
  pg_stat_have_stats 
 --------------------
  t
@@ -1233,7 +1344,7 @@ SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
 
 REINDEX index CONCURRENTLY stats_test_idx1;
 -- false for previous oid
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
  pg_stat_have_stats 
 --------------------
  f
@@ -1241,7 +1352,7 @@ SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
 
 -- true for new oid
 SELECT 'stats_test_idx1'::regclass::oid AS stats_test_idx1_oid \gset
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
  pg_stat_have_stats 
 --------------------
  t
@@ -1249,7 +1360,7 @@ SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
 
 -- pg_stat_have_stats returns true for a rolled back drop index with stats
 BEGIN;
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
  pg_stat_have_stats 
 --------------------
  t
@@ -1257,7 +1368,7 @@ SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
 
 DROP index stats_test_idx1;
 ROLLBACK;
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
  pg_stat_have_stats 
 --------------------
  t
@@ -1578,7 +1689,7 @@ DECLARE
 BEGIN
   -- we don't want to wait forever; loop will exit after 30 seconds
   FOR i IN 1 .. 300 LOOP
-    SELECT (pg_stat_get_tuples_hot_updated('brin_hot'::regclass::oid) > 0) INTO updated;
+    SELECT (pg_stat_get_tab_tuples_hot_updated('brin_hot'::regclass::oid) > 0) INTO updated;
     EXIT WHEN updated;
 
     -- wait a little
@@ -1605,10 +1716,10 @@ SELECT wait_for_hot_stats();
  
 (1 row)
 
-SELECT pg_stat_get_tuples_hot_updated('brin_hot'::regclass::oid);
- pg_stat_get_tuples_hot_updated 
---------------------------------
-                              1
+SELECT pg_stat_get_tab_tuples_hot_updated('brin_hot'::regclass::oid);
+ pg_stat_get_tab_tuples_hot_updated 
+------------------------------------
+                                  1
 (1 row)
 
 DROP TABLE brin_hot;
diff --git a/src/test/regress/sql/stats.sql b/src/test/regress/sql/stats.sql
index 13db45d4dc..34a9a9378e 100644
--- a/src/test/regress/sql/stats.sql
+++ b/src/test/regress/sql/stats.sql
@@ -126,6 +126,77 @@ FROM prevstats AS pr;
 
 COMMIT;
 
+----
+-- Basic tests for toast
+---
+CREATE TABLE stats_toast(stuff text);
+ALTER TABLE stats_toast ALTER COLUMN stuff SET STORAGE EXTERNAL;
+INSERT INTO stats_toast VALUES (repeat('a',1000000));
+SELECT count(*) FROM stats_toast WHERE stuff like '%a%';
+
+-- ensure pending stats are flushed
+SELECT pg_stat_force_next_flush();
+
+select toast_blks_read >= 0 as toast_blks_read, toast_blks_hit > 0 as toast_blks_hit,
+       tidx_blks_read >= 0 as tidx_blks_read, tidx_blks_hit > 0 as tidx_blks_hit,
+       toast_blks_hit >= toast_blks_read,
+       tidx_blks_hit >= tidx_blks_read
+   from pg_statio_all_tables
+ where relname = 'stats_toast';
+
+----
+-- Basic tests for partition
+---
+CREATE TABLE stats_partition (
+    id1         int not null,
+    id2         int not null
+) PARTITION BY RANGE (id2);
+
+CREATE TABLE stats_partition_5 PARTITION OF stats_partition
+    FOR VALUES FROM (0) TO (5);
+
+CREATE TABLE stats_partition_20000 PARTITION OF stats_partition
+    FOR VALUES FROM (5) TO (20001);
+
+CREATE INDEX ON stats_partition (id2);
+
+insert into stats_partition select a,a from generate_series(0,20000) a;
+
+SET enable_seqscan TO off;
+SET enable_bitmapscan TO off;
+
+select * from stats_partition where id2 = 2000;
+select * from stats_partition where id2 = 2;
+-- ensure pending stats are flushed
+SELECT pg_stat_force_next_flush();
+
+select relname, indexrelname, idx_scan > 0 as idx_scan, idx_tup_read > 0 as idx_tup_read,
+       idx_tup_fetch > 0 as idx_tup_fetch
+   from pg_stat_all_indexes
+ where relname like '%stats_partition%'
+ order by relname COLLATE "C";
+
+select relname, indexrelname, idx_blks_read >= 0 as idx_blks_read,
+       idx_blks_hit > 0 as idx_blks_hit
+   from pg_statio_all_indexes
+ where relname like '%stats_partition%'
+ order by relname COLLATE "C";
+
+select relname,seq_scan > 0 as seq_scan, seq_tup_read > 0 as seq_tup_read,
+       idx_scan > 0 as idx_scan, idx_tup_fetch > 0 as idx_tup_fetch
+   from pg_stat_all_tables
+  where relname like '%stats_partition%'
+  order by relname COLLATE "C";
+
+select relname, heap_blks_read >= 0 as heap_blks_read, heap_blks_hit > 0 as heap_blks_hit,
+       idx_blks_read >= 0 as idx_blks_read, idx_blks_hit > 0 as idx_blks_hit
+   from pg_statio_all_tables
+ where relname like '%stats_partition%'
+ order by relname COLLATE "C";
+
+SET enable_seqscan TO on;
+SET enable_bitmapscan TO on;
+
 ----
 -- Basic tests for track_functions
 ---
@@ -218,73 +289,73 @@ SELECT 'drop_stats_test_subxact'::regclass::oid AS drop_stats_test_subxact_oid \
 
 SELECT pg_stat_force_next_flush();
 
-SELECT pg_stat_get_live_tuples(:drop_stats_test_oid);
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_oid);
 DROP TABLE drop_stats_test;
-SELECT pg_stat_get_live_tuples(:drop_stats_test_oid);
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_oid);
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_oid);
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_oid);
 
 -- check that rollback protects against having stats dropped and that local
 -- modifications don't pose a problem
-SELECT pg_stat_get_live_tuples(:drop_stats_test_xact_oid);
-SELECT pg_stat_get_tuples_inserted(:drop_stats_test_xact_oid);
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_xact_oid);
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_xact_oid);
+SELECT pg_stat_get_tab_tuples_inserted(:drop_stats_test_xact_oid);
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_xact_oid);
 BEGIN;
 INSERT INTO drop_stats_test_xact DEFAULT VALUES;
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_xact_oid);
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_xact_oid);
 DROP TABLE drop_stats_test_xact;
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_xact_oid);
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_xact_oid);
 ROLLBACK;
 SELECT pg_stat_force_next_flush();
-SELECT pg_stat_get_live_tuples(:drop_stats_test_xact_oid);
-SELECT pg_stat_get_tuples_inserted(:drop_stats_test_xact_oid);
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_xact_oid);
+SELECT pg_stat_get_tab_tuples_inserted(:drop_stats_test_xact_oid);
 
 -- transactional drop
-SELECT pg_stat_get_live_tuples(:drop_stats_test_xact_oid);
-SELECT pg_stat_get_tuples_inserted(:drop_stats_test_xact_oid);
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_xact_oid);
+SELECT pg_stat_get_tab_tuples_inserted(:drop_stats_test_xact_oid);
 BEGIN;
 INSERT INTO drop_stats_test_xact DEFAULT VALUES;
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_xact_oid);
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_xact_oid);
 DROP TABLE drop_stats_test_xact;
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_xact_oid);
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_xact_oid);
 COMMIT;
 SELECT pg_stat_force_next_flush();
-SELECT pg_stat_get_live_tuples(:drop_stats_test_xact_oid);
-SELECT pg_stat_get_tuples_inserted(:drop_stats_test_xact_oid);
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_xact_oid);
+SELECT pg_stat_get_tab_tuples_inserted(:drop_stats_test_xact_oid);
 
 -- savepoint rollback (2 levels)
-SELECT pg_stat_get_live_tuples(:drop_stats_test_subxact_oid);
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_subxact_oid);
 BEGIN;
 INSERT INTO drop_stats_test_subxact DEFAULT VALUES;
 SAVEPOINT sp1;
 INSERT INTO drop_stats_test_subxact DEFAULT VALUES;
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_subxact_oid);
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_subxact_oid);
 SAVEPOINT sp2;
 DROP TABLE drop_stats_test_subxact;
 ROLLBACK TO SAVEPOINT sp2;
-SELECT pg_stat_get_xact_tuples_inserted(:drop_stats_test_subxact_oid);
+SELECT pg_stat_get_tab_xact_tuples_inserted(:drop_stats_test_subxact_oid);
 COMMIT;
 SELECT pg_stat_force_next_flush();
-SELECT pg_stat_get_live_tuples(:drop_stats_test_subxact_oid);
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_subxact_oid);
 
 -- savepoint rolback (1 level)
-SELECT pg_stat_get_live_tuples(:drop_stats_test_subxact_oid);
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_subxact_oid);
 BEGIN;
 SAVEPOINT sp1;
 DROP TABLE drop_stats_test_subxact;
 SAVEPOINT sp2;
 ROLLBACK TO SAVEPOINT sp1;
 COMMIT;
-SELECT pg_stat_get_live_tuples(:drop_stats_test_subxact_oid);
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_subxact_oid);
 
 -- and now actually drop
-SELECT pg_stat_get_live_tuples(:drop_stats_test_subxact_oid);
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_subxact_oid);
 BEGIN;
 SAVEPOINT sp1;
 DROP TABLE drop_stats_test_subxact;
 SAVEPOINT sp2;
 RELEASE SAVEPOINT sp1;
 COMMIT;
-SELECT pg_stat_get_live_tuples(:drop_stats_test_subxact_oid);
+SELECT pg_stat_get_tab_live_tuples(:drop_stats_test_subxact_oid);
 
 DROP TABLE trunc_stats_test, trunc_stats_test1, trunc_stats_test2, trunc_stats_test3, trunc_stats_test4;
 DROP TABLE prevstats;
@@ -561,40 +632,40 @@ CREATE index stats_test_idx1 on stats_test_tab1(a);
 SELECT 'stats_test_idx1'::regclass::oid AS stats_test_idx1_oid \gset
 SET enable_seqscan TO off;
 select a from stats_test_tab1 where a = 3;
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
 
 -- pg_stat_have_stats returns false for dropped index with stats
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
 DROP index stats_test_idx1;
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
 
 -- pg_stat_have_stats returns false for rolled back index creation
 BEGIN;
 CREATE index stats_test_idx1 on stats_test_tab1(a);
 SELECT 'stats_test_idx1'::regclass::oid AS stats_test_idx1_oid \gset
 select a from stats_test_tab1 where a = 3;
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
 ROLLBACK;
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
 
 -- pg_stat_have_stats returns true for reindex CONCURRENTLY
 CREATE index stats_test_idx1 on stats_test_tab1(a);
 SELECT 'stats_test_idx1'::regclass::oid AS stats_test_idx1_oid \gset
 select a from stats_test_tab1 where a = 3;
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
 REINDEX index CONCURRENTLY stats_test_idx1;
 -- false for previous oid
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
 -- true for new oid
 SELECT 'stats_test_idx1'::regclass::oid AS stats_test_idx1_oid \gset
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
 
 -- pg_stat_have_stats returns true for a rolled back drop index with stats
 BEGIN;
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
 DROP index stats_test_idx1;
 ROLLBACK;
-SELECT pg_stat_have_stats('relation', :dboid, :stats_test_idx1_oid);
+SELECT pg_stat_have_stats('index', :dboid, :stats_test_idx1_oid);
 
 -- put enable_seqscan back to on
 SET enable_seqscan TO on;
@@ -793,7 +864,7 @@ DECLARE
 BEGIN
   -- we don't want to wait forever; loop will exit after 30 seconds
   FOR i IN 1 .. 300 LOOP
-    SELECT (pg_stat_get_tuples_hot_updated('brin_hot'::regclass::oid) > 0) INTO updated;
+    SELECT (pg_stat_get_tab_tuples_hot_updated('brin_hot'::regclass::oid) > 0) INTO updated;
     EXIT WHEN updated;
 
     -- wait a little
@@ -818,7 +889,7 @@ UPDATE brin_hot SET val = -3 WHERE id = 42;
 \c -
 
 SELECT wait_for_hot_stats();
-SELECT pg_stat_get_tuples_hot_updated('brin_hot'::regclass::oid);
+SELECT pg_stat_get_tab_tuples_hot_updated('brin_hot'::regclass::oid);
 
 DROP TABLE brin_hot;
 DROP FUNCTION wait_for_hot_stats();
diff --git a/src/tools/pgindent/typedefs.list b/src/tools/pgindent/typedefs.list
index bf50a32119..10caf1706a 100644
--- a/src/tools/pgindent/typedefs.list
+++ b/src/tools/pgindent/typedefs.list
@@ -2042,10 +2042,11 @@ PgStatShared_Database
 PgStatShared_Function
 PgStatShared_HashEntry
 PgStatShared_IO
-PgStatShared_Relation
+PgStatShared_Index
 PgStatShared_ReplSlot
 PgStatShared_SLRU
 PgStatShared_Subscription
+PgStatShared_Table
 PgStatShared_Wal
 PgStat_ArchiverStats
 PgStat_BackendSubEntry
@@ -2059,6 +2060,7 @@ PgStat_FetchConsistency
 PgStat_FunctionCallUsage
 PgStat_FunctionCounts
 PgStat_HashKey
+PgStat_IndexStatus
 PgStat_IO
 PgStat_Kind
 PgStat_KindInfo
-- 
2.34.1

